#! python
# -*- coding: utf-8 -*-

import glob
import logging
import os
import xml.etree.ElementTree as ET
from collections import Counter
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from gseapy.base import GSEAbase
from gseapy.gse import Metric, gsea_rs, prerank2d_rs, prerank_rs
from gseapy.parser import gsea_cls_parser
from gseapy.plot import gseaplot

# from memory_profiler import profile


class GSEA(GSEAbase):
    """GSEA main tool"""

    def __init__(
        self,
        data: Union[pd.DataFrame, str],
        gene_sets: Union[List[str], str, Dict[str, str]],
        classes: Union[List[str], str, Dict[str, str]],
        outdir: Optional[str] = None,
        min_size: int = 15,
        max_size: int = 500,
        permutation_num: int = 1000,
        weight: float = 1.0,
        permutation_type: str = "phenotype",
        method: str = "signal_to_noise",
        ascending: bool = False,
        threads: int = 1,
        figsize: Tuple[float, float] = (6.5, 6),
        format: str = "pdf",
        graph_num: int = 20,
        no_plot: bool = False,
        seed: int = 123,
        verbose: bool = False,
    ):
        """
        GSEA main tool.

        Parameters
        ----------
        data : pandas.DataFrame or str
            Data table with genes in the index and samples in the columns.
            If str, the file path to the data file (e.g. .txt, .csv, .tsv).
        gene_sets : list or str or dict
            Gene sets file path or list of gene sets.
            If str, the file path to the gene sets file (e.g. .gmt, .txt, .csv).
            If dict, a dictionary of gene sets.
        classes : list or str or dict
            Phenotype labels file path or list of phenotype labels.
            If str, the file path to the phenotype labels file (e.g. .txt, .csv, .tsv).
            If dict, a dictionary of phenotype labels.
        outdir : str, optional
            Output directory. If not provided, the output will be saved.
        min_size : int, optional
            Minimum number of genes in a gene set. Default is 15.
        max_size : int, optional
            Maximum number of genes in a gene set. Default is 500.
        permutation_num : int, optional
            Number of permutations. Default is 1000.
        weight : float, optional
            Weighting factor used in the calculation of the ES. Default is 1.0.
        permutation_type : str, optional
            Type of permutation, either 'phenotype' or 'gene_set'. Default is 'phenotype'.
        method : str, optional
            Metric used in the calculation of the ES. Default is 'signal_to_noise'.
        ascending : bool, optional
            If True, the genes will be ranked in ascending order. Default is False.
        threads : int, optional
            Number of threads to use. Default is 1.
        figsize : tuple, optional
            Figure size for the enrichment plot. Default is (6.5, 6).
        format : str, optional
            Format of the output file. Default is 'pdf'.
        graph_num : int, optional
            Number of enrichment plots to generate. Default is 20.
        no_plot : bool, optional
            If True, no enrichment plots will be generated. Default is False.
        seed : int, optional
            Random seed for the permutation test. Default is 123.
        verbose : bool, optional
            If True, the progress bar will be shown. Default is False.

        Examples
        --------
        >>> from gseapy import GSEA
        >>> gsea = GSEA(data='expression.txt', gene_sets='gene_sets.gmt', classes='test.cls')
        >>> gsea.run()
        """
        super(GSEA, self).__init__(
            outdir=outdir,
            gene_sets=gene_sets,
            module="gsea",
            threads=threads,
            verbose=verbose,
        )
        self.data = data
        # self.classes = classes
        self.permutation_type = permutation_type
        self.method = method
        self.min_size = min_size
        self.max_size = max_size
        self.permutation_num = int(permutation_num) if int(permutation_num) > 0 else 0
        self.weight = weight
        self.ascending = ascending
        self.figsize = figsize
        self.format = format
        self.graph_num = int(graph_num)
        self.seed = seed
        self.ranking = None
        self._noplot = no_plot
        # some preprocessing
        assert self.permutation_type in ["phenotype", "gene_set"]
        assert self.min_size <= self.max_size
        assert self.min_size > 0
        # phenotype labels parsing
        self.load_classes(classes)

    def load_data(self) -> Tuple[pd.DataFrame, Dict]:
        """pre-processed the data frame.new filtering methods will be implement here."""
        exprs = self._load_data(self.data)
        exprs = self._check_data(exprs)
        exprs, cls_dict = self._filter_data(exprs)

        return exprs, cls_dict

    def _map_classes(self, sample_names: List[str]) -> Dict[str, Any]:
        """
        update
        """
        cls_dict = self.groups
        if isinstance(self.groups, dict):
            # update groups
            self.groups = [cls_dict[c] for c in sample_names]
        else:
            cls_dict = {k: v for k, v in zip(sample_names, self.groups)}
        return cls_dict

    def _filter_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        filter data rows with std == 0
        """
        # in case the description column is numeric
        if len(self.groups) == (df.shape[1] - 1):
            df = df.iloc[:, 1:]
        cls_dict = self._map_classes(df.columns)
        # remove rows that are all zeros !
        df = df.loc[df.abs().sum(axis=1) > 0, :]
        # drop gene which std == 0 in all samples
        # compatible to py3.7
        major, minor, _ = [int(i) for i in pd.__version__.split(".")]
        # handle cases for samples < 3, use mean
        if (major == 1 and minor < 5) or (major < 1):
            # fix numeric_only error
            df_std = df.groupby(by=cls_dict, axis=1).std(ddof=0)
        else:
            gene_idxs = df.index.to_list()
            # df_std = df.groupby(by=cls_dict, axis=1).std(numeric_only=True, ddof=0)
            df_std = (
                df.T.groupby(by=cls_dict)[gene_idxs].std(numeric_only=True, ddof=0).T
            )

        # remove rows that std are zeros for sample size >= 3 in each group
        if all(map(lambda a: a[1] >= 3, Counter(cls_dict.values()).items())):
            df = df[df_std.abs().sum(axis=1) > 0]
        df = df + 1e-08  # we don't like zeros in denominator !!!
        # data frame must have length > 1
        assert df.shape[0] > 1

        return df, cls_dict

    def calc_metric(
        self,
        df: pd.DataFrame,
        method: str,
        pos: str,
        neg: str,
        classes: Dict[str, str],
        ascending: bool,
    ) -> Tuple[List[int], pd.Series]:
        """The main function to rank an expression table. works for 2d array.

        :param df:      gene_expression DataFrame.
        :param method:  The method used to calculate a correlation or ranking. Default: 'log2_ratio_of_classes'.
                        Others methods are:

                        1. 'signal_to_noise' (s2n) or 'abs_signal_to_noise' (abs_s2n)

                            You must have at least three samples for each phenotype.
                            The more distinct the gene expression is in each phenotype,
                            the more the gene acts as a “class marker”.

                        2. 't_test'

                            Uses the difference of means scaled by the standard deviation and number of samples.
                            Note: You must have at least three samples for each phenotype to use this metric.
                            The larger the t-test ratio, the more distinct the gene expression is in each phenotype
                            and the more the gene acts as a “class marker.”

                        3. 'ratio_of_classes' (also referred to as fold change).

                            Uses the ratio of class means to calculate fold change for natural scale data.

                        4. 'diff_of_classes'

                            Uses the difference of class means to calculate fold change for natural scale data

                        5. 'log2_ratio_of_classes'

                            Uses the log2 ratio of class means to calculate fold change for natural scale data.
                            This is the recommended statistic for calculating fold change for log scale data.

        :param str pos: one of labels of phenotype's names.
        :param str neg: one of labels of phenotype's names.
        :param dict classes: column id to group mapping.
        :param bool ascending:  bool or list of bool. Sort ascending vs. descending.
        :return: returns argsort values of a tuple where
            0: argsort positions (indices)
            1: pd.Series of correlation value. Gene_name is index, and value is rankings.

        visit here for more docs: http://software.broadinstitute.org/gsea/doc/GSEAUserGuideFrame.html
        """

        # exclude any zero stds.
        # compatible to py3.7
        major, minor, _ = [int(i) for i in pd.__version__.split(".")]
        if (major == 1 and minor < 5) or (major < 1):
            # fix numeric_only error
            df_mean = df.groupby(by=classes, axis=1).mean()
            df_std = df.groupby(by=classes, axis=1).std()
        else:
            gene_idxs = df.index.to_list()
            df_std = df.T.groupby(by=classes)[gene_idxs].std(numeric_only=True).T
            df_mean = df.T.groupby(by=classes)[gene_idxs].mean(numeric_only=True).T
            # df_mean = df.groupby(by=classes, axis=1).mean(numeric_only=True)
            # df_std = df.groupby(by=classes, axis=1).std(numeric_only=True)
        class_values = Counter(classes.values())
        n_pos = class_values[pos]
        n_neg = class_values[neg]

        if method in ["signal_to_noise", "s2n"]:
            ser = (df_mean[pos] - df_mean[neg]) / (df_std[pos] + df_std[neg])
        elif method in ["abs_signal_to_noise", "abs_s2n"]:
            ser = ((df_mean[pos] - df_mean[neg]) / (df_std[pos] + df_std[neg])).abs()
        elif method == "t_test":
            ser = (df_mean[pos] - df_mean[neg]) / np.sqrt(
                df_std[pos] ** 2 / n_pos + df_std[neg] ** 2 / n_neg
            )
        elif method == "ratio_of_classes":
            ser = df_mean[pos] / df_mean[neg]
        elif method == "diff_of_classes":
            ser = df_mean[pos] - df_mean[neg]
        elif method == "log2_ratio_of_classes":
            ser = np.log2(df_mean[pos] / df_mean[neg])
            if ser.isna().sum() > 0:
                self._logger.warning("Invalid value encountered in log2, and dumped.")
                ser = ser.dropna()
                assert len(ser) > 1
        else:
            logging.error("Please provide correct method name!!!")
            raise LookupError("Input method: %s is not supported" % method)
        ser_ind = ser.values.argsort().tolist()
        ser = ser.iloc[ser_ind]
        if ascending:
            return ser_ind, ser
        # descending order
        return ser_ind[::-1], ser[::-1]

    def _check_classes(self, counter: Counter) -> List[str]:
        """
        check each cls group length
        """
        # metrics = ["signal_to_noise", "s2n", "abs_signal_to_noise", "abs_s2n", "t_test"]
        s = []
        for c, v in sorted(counter.items(), key=lambda item: item[1]):
            if v < 3:
                if self.permutation_type == "phenotype":
                    self._logger.warning(
                        f"Number of {c}: {v}, it must be >= 3 for permutation type: phenotype !"
                    )
                    self._logger.warning("Permutation type change to gene_set.")
                    self.permutation_type == "gene_set"
            s.append(c)
        return s

    def load_classes(self, classes: Union[str, List[str], Dict[str, Any]]):
        """Parse group (classes)"""
        if isinstance(classes, dict):
            # check number of samples
            s = self._check_classes(Counter(classes.values()))
            self.pheno_pos = s[0]
            self.pheno_neg = s[1]
            # n_pos = class_values[pos]
            # n_neg = class_values[neg]
            self.groups = classes
        else:
            pos, neg, cls_vector = gsea_cls_parser(classes)
            s = self._check_classes(Counter(cls_vector))
            self.pheno_pos = pos
            self.pheno_neg = neg
            self.groups = cls_vector

    def to_cls(self, outdir: str):
        """Save group information to cls file"""
        with open(os.path.join(outdir, "group.cls"), "w") as f:
            f.write(f"{len(self.groups)} 2 1\n")
            f.write(f"# {self.pheno_pos} {self.pheno_neg}\n")
            if isinstance(self.groups, dict):
                f.write(" ".join(list(self.groups.values())) + "\n")
            else:
                f.write(" ".join(self.groups) + "\n")

    # @profile
    def run(self):
        """GSEA main procedure"""
        m = self.method.lower()
        if m in ["signal_to_noise", "s2n"]:
            method = Metric.Signal2Noise
        elif m in ["abs_signal_to_noise", "abs_s2n"]:
            method = Metric.AbsSignal2Noise
        elif m == "t_test":
            method = Metric.Ttest
        elif m == "ratio_of_classes":
            method = Metric.RatioOfClasses
        elif m == "diff_of_classes":
            method = Metric.DiffOfClasses
        elif m == "log2_ratio_of_classes":
            method = Metric.Log2RatioOfClasses
        else:
            raise Exception("Sorry, input method %s is not supported" % m)

        # Start Analysis
        self._logger.info("Parsing data files for GSEA.............................")
        # select correct expression genes and values.
        dat, cls_dict = self.load_data()
        # filtering out gene sets and build gene sets dictionary
        self._gene_isupper = self.check_uppercase(gene_list=dat.index.values)
        gmt = self.load_gmt(gene_list=dat.index.values, gmt=self.gene_sets)
        self.gmt = gmt
        self._logger.info(
            "%04d gene_sets used for further statistical testing....." % len(gmt)
        )
        self._logger.info("Start to run GSEA...Might take a while..................")
        # gene symbols
        gene_names = dat.index.to_list()
        if (not self._gene_isupper) and self._gene_toupper:
            gene_names = [str(x).upper() for x in gene_names]
            dat.index = gene_names
            self._logger.info("Genes are converted to uppercase.")

        # compute ES, NES, pval, FDR, RES
        if self.permutation_type == "gene_set":
            # ranking metrics calculation.
            idx, dat2 = self.calc_metric(
                df=dat,
                method=self.method,
                pos=self.pheno_pos,
                neg=self.pheno_neg,
                classes=cls_dict,
                ascending=self.ascending,
            )
            gsum = prerank_rs(
                dat2.index.to_list(),  # gene list
                dat2.squeeze().values.tolist(),  # ranking values
                gmt,  # must be a dict object
                self.weight,
                self.min_size,
                self.max_size,
                self.permutation_num,
                self._threads,
                self.seed,
            )
            ## need to update indices, prerank_rs only stores input's order
            # so compatible with code code below
            indices = gsum.indices
            indices[0] = idx
            gsum.indices = indices  # only accept [[]]
        else:  # phenotype permutation
            group = list(
                map(lambda x: True if x == self.pheno_pos else False, self.groups)
            )
            gsum = gsea_rs(
                dat.index.to_list(),
                dat.values.tolist(),  # each row is gene values across samples
                gmt,
                group,
                method,
                self.weight,
                self.min_size,
                self.max_size,
                self.permutation_num,
                self._threads,
                self.seed,
            )

        if self._outdir is not None:
            self._logger.info(
                "Start to generate GSEApy reports and figures............"
            )
        self.ranking = pd.Series(gsum.rankings[0], index=dat.index[gsum.indices[0]])
        # reorder datarame for heatmap
        # self._heatmat(df=dat.loc[dat2.index], classes=cls_vector)
        self._heatmat(df=dat.iloc[gsum.indices[0]], classes=self.groups)
        # write output and plotting
        self.to_df(gsum.summaries, gmt, self.ranking)
        if self._outdir is not None:
            self.to_cls(outdir=self.outdir)
            self.ranking.to_csv(
                os.path.join(self.outdir, "gsea_data.rnk"), sep="\t", header=False
            )
        self._logger.info("Congratulations. GSEApy ran successfully.................\n")

        return


class Prerank(GSEAbase):
    """GSEA prerank tool"""

    def __init__(
        self,
        rnk: Union[pd.DataFrame, pd.Series, str],
        gene_sets: Union[List[str], str, Dict[str, str]],
        outdir: Optional[str] = None,
        pheno_pos="Pos",
        pheno_neg="Neg",
        min_size: int = 15,
        max_size: int = 500,
        permutation_num: int = 1000,
        weight: float = 1.0,
        ascending: Optional[bool] = False,
        threads: int = 1,
        figsize: Tuple[float, float] = (6.5, 6),
        format: str = "pdf",
        graph_num: int = 20,
        no_plot: bool = False,
        seed: int = 123,
        verbose: bool = False,
    ):
        super(Prerank, self).__init__(
            outdir=outdir,
            gene_sets=gene_sets,
            module="prerank",
            threads=threads,
            verbose=verbose,
        )
        self.rnk = rnk
        self.pheno_pos = pheno_pos
        self.pheno_neg = pheno_neg
        self.min_size = min_size
        self.max_size = max_size
        self.permutation_num = int(permutation_num) if int(permutation_num) > 0 else 0
        self.weight = weight
        self.ascending = ascending
        self.figsize = figsize
        self.format = format
        self.graph_num = int(graph_num)
        self.seed = seed
        self.ranking = None
        self._noplot = no_plot
        self.permutation_type = "gene_set"
        assert self.min_size > 0

    def _load_ranking(self, rank_metric: pd.DataFrame) -> pd.Series:
        """Parse ranking
        rank_metric: two column dataframe. first column is gene ids

        """
        # load data
        rnk_cols = rank_metric.columns
        # if case the input has more than two columns, only select the last two
        # this happens when gene id is index, while gene name and ranking value are columns
        if len(rnk_cols) > 2:
            rank_metric = rank_metric.iloc[:, -2:]
            rnk_cols = rank_metric.columns
        # if self.ascending is None and not rank_metric.iloc[:, -1].is_monotonic_decreasing:
        #     rank_metric.sort_values(by=rnk_cols[-1], ascending=False, inplace=True)
        if self.ascending is not None:
            rank_metric.sort_values(
                by=rnk_cols[-1], ascending=self.ascending, inplace=True
            )
        # drop na values
        if rank_metric.isnull().any(axis=1).sum() > 0:
            self._logger.warning(
                "Input gene rankings contains NA values(gene name and ranking value), drop them all!"
            )
            # print out NAs
            NAs = rank_metric[rank_metric.isnull().any(axis=1)]
            self._logger.debug("NAs list:\n" + NAs.to_string())
            rank_metric.dropna(how="any", inplace=True)
        # rename duplicate id, make them unique
        rank_metric = self.make_unique(rank_metric, col_idx=0)
        # reset ranking index, because you have sort values and drop duplicates.
        rank_metric.reset_index(drop=True, inplace=True)
        rank_metric.columns = ["gene_name", "prerank"]
        rankser = rank_metric.set_index("gene_name", drop=True).squeeze()

        # check whether contains infinity values
        if np.isinf(rankser).values.sum() > 0:
            self._logger.warning("Input gene rankings contains inf values!")
            rankser.replace(-np.inf, method="ffill", inplace=True)
            rankser.replace(np.inf, method="bfill", inplace=True)

        # check duplicate values and warning
        dups = rankser.duplicated().sum()
        if dups > 0:
            msg = (
                "Duplicated values found in preranked stats: {:.2%} of genes\n".format(
                    dups / rankser.size
                )
            )
            msg += "The order of those genes will be arbitrary, which may produce unexpected results."
            self._logger.warning(msg)

        # return series
        return rankser

    def load_ranking(self):
        """
        parse rnk input
        """
        rank_metric = self._load_data(self.rnk)  # gene id is the first column
        if rank_metric.select_dtypes(np.number).shape[1] == 1:
            # return series
            return self._load_ranking(rank_metric)
        ## In case the input type multi-column ranking dataframe
        # drop na gene id values
        rank_metric = rank_metric.dropna(subset=rank_metric.columns[0])
        # make unique
        rank_metric = self.make_unique(rank_metric, col_idx=0)
        # set index
        rank_metric = self._check_data(rank_metric)
        # check ties in prerank stats
        dups = rank_metric.apply(lambda df: df.duplicated().sum() / df.size)
        if (dups > 0).sum() > 0:
            msg = "Duplicated values found in preranked stats:\nsample\tratio\n%s\n" % (
                dups.to_string(float_format="{:,.2%}".format)
            )
            msg += "The order of those genes will be arbitrary, which may produce unexpected results."
            self._logger.warning(msg)

        return rank_metric

    # @profile
    def run(self):
        """GSEA prerank workflow"""

        assert self.min_size <= self.max_size

        # parsing rankings
        dat2 = self.load_ranking()
        assert len(dat2) > 1
        self.ranking = dat2
        # Start Analysis
        self._logger.info("Parsing data files for GSEA.............................")
        # filtering out gene sets and build gene sets dictionary
        self._gene_isupper = self.check_uppercase(gene_list=dat2.index.values)
        gmt = self.load_gmt(gene_list=dat2.index.values, gmt=self.gene_sets)
        self.gmt = gmt
        self._logger.info(
            "%04d gene_sets used for further statistical testing....." % len(gmt)
        )
        self._logger.info("Start to run GSEA...Might take a while..................")
        gene_names = dat2.index.to_list()
        if (not self._gene_isupper) and self._gene_toupper:
            gene_names = [str(x).upper() for x in gene_names]
            self._logger.info("Genes are converted to uppercase.")
        # compute ES, NES, pval, FDR, RES
        if isinstance(dat2, pd.DataFrame):
            _prerank = prerank2d_rs
        else:
            _prerank = prerank_rs
        # run
        gsum = _prerank(
            gene_names,  # gene list
            dat2.values.tolist(),  # ranking values
            gmt,  # must be a dict object
            self.weight,
            self.min_size,
            self.max_size,
            self.permutation_num,
            self._threads,
            self.seed,
        )
        self.to_df(
            gsea_summary=gsum.summaries,
            gmt=gmt,
            rank_metric=dat2,
            indices=gsum.indices if isinstance(dat2, pd.DataFrame) else None,
        )
        if self._outdir is not None:
            self.ranking.to_csv(
                os.path.join(self.outdir, "prerank_data.rnk"), sep="\t", header=False
            )

        self._logger.info("Congratulations. GSEApy runs successfully................\n")

        return


class Replot(GSEAbase):
    """To reproduce GSEA desktop output results."""

    def __init__(
        self,
        indir: str,
        outdir: str = "GSEApy_Replot",
        weight: float = 1.0,
        min_size: int = 3,
        max_size: int = 1000,
        figsize: Tuple[float, float] = (6.5, 6),
        format: str = "pdf",
        verbose: bool = False,
    ):
        self.indir = indir
        self.outdir = outdir
        self.weight = weight
        self.min_size = min_size
        self.max_size = max_size
        self.figsize = figsize
        self.format = format
        self.verbose = bool(verbose)
        self.module = "replot"
        self.gene_sets = None
        self.ascending = False
        # init logger
        self.prepare_outdir()

    def gsea_edb_parser(self, results_path):
        """Parse results.edb file stored under **edb** file folder.

        :param results_path: the path of results.edb file.
        :return:
            a dict contains { enrichment_term: [es, nes, pval, fdr, fwer, hit_ind]}
        """

        xtree = ET.parse(results_path)
        xroot = xtree.getroot()
        res = {}
        # dict_keys(['RANKED_LIST', 'GENESET', 'FWER', 'ES_PROFILE',
        # 'HIT_INDICES', 'ES', 'NES', 'TEMPLATE', 'RND_ES', 'RANK_SCORE_AT_ES',
        # 'NP', 'RANK_AT_ES', 'FDR'])
        for node in xroot.findall("DTG"):
            enrich_term = node.attrib.get("GENESET").split("#")[1]
            es_profile = node.attrib.get("ES_PROFILE").split(" ")
            # esnull = term.get('RND_ES').split(" ")
            hit_ind = node.attrib.get("HIT_INDICES").split(" ")
            es_profile = [float(i) for i in es_profile]
            hit_ind = [int(i) for i in hit_ind]
            # esnull = [float(i) for i in esnull ]
            es = float(node.attrib.get("ES"))
            nes = float(node.attrib.get("NES"))
            pval = float(node.attrib.get("NP"))
            fdr = float(node.attrib.get("FDR"))
            fwer = float(node.attrib.get("FWER"))
            logging.debug("Enriched Gene set is: " + enrich_term)
            res[enrich_term] = [es, nes, pval, fdr, fwer, hit_ind]
        return res

    def run(self):
        """main replot function"""
        assert self.min_size <= self.max_size

        # parsing files.......
        try:
            results_path = glob.glob(os.path.join(self.indir, "edb/results.edb"))[0]
            rank_path = glob.glob(os.path.join(self.indir, "edb/*.rnk"))[0]
            gene_set_path = glob.glob(os.path.join(self.indir, "edb/gene_sets.gmt"))[0]
        except IndexError:
            raise Exception("Could not locate GSEA files in the given directory!")
        # extract sample names from .cls file
        cls_path = glob.glob(os.path.join(self.indir, "*/edb/*.cls"))
        if cls_path:
            pos, neg, classes = gsea_cls_parser(cls_path[0])
        else:
            # logic for prerank results
            pos, neg = "", ""
        # start reploting
        self.gene_sets = gene_set_path
        # obtain gene sets
        gene_set_dict = self.parse_gmt(gmt=gene_set_path)
        # obtain rank_metrics
        rank_metric = self._load_data(rank_path)
        # rank_metric = rank_metric.set_index(rank_metric.columns[0])
        correl_vector = rank_metric.iloc[:, 1].values
        gene_list = rank_metric.iloc[:, 0].values
        # extract each enriment term in the results.edb files and plot.

        database = self.gsea_edb_parser(results_path)
        for enrich_term, data in database.items():
            # extract statistical resutls from results.edb file
            es, nes, pval, fdr, fwer, hit_ind = data
            gene_set = gene_set_dict.get(enrich_term)
            # calculate enrichment score
            RES = self.enrichment_score(
                gene_list=gene_list,
                correl_vector=correl_vector,
                gene_set=gene_set,
                weight=self.weight,
                nperm=0,
            )[-1]
            # plotting
            term = enrich_term.replace("/", "_").replace(":", "_")
            outfile = "{0}/{1}.{2}.{3}".format(
                self.outdir, term, self.module, self.format
            )
            gseaplot(
                rank_metric=correl_vector,
                term=enrich_term,
                hits=hit_ind,
                nes=nes,
                pval=pval,
                fdr=fdr,
                RES=RES,
                pheno_pos=pos,
                pheno_neg=neg,
                figsize=self.figsize,
                ofname=outfile,
            )

        self._logger.info(
            "Congratulations! Your plots have been reproduced successfully!\n"
        )



#! python
# -*- coding: utf-8 -*-

from typing import Dict, List, Optional, Union

import pandas as pd

from gseapy.base import GSEAbase
from gseapy.gse import gsva_rs
from gseapy.utils import mkdirs


class GSVA(GSEAbase):
    """GSVA"""

    def __init__(
        self,
        data: Union[pd.DataFrame, pd.Series, str],
        gene_sets: Union[List[str], str, Dict[str, str]],
        outdir: Optional[str] = None,
        kcdf: Optional[str] = "Gaussian",
        weight: float = 1.0,
        mx_diff: bool = True,
        abs_rnk: bool = False,
        min_size: int = 15,
        max_size: int = 500,
        threads: int = 1,
        seed: int = 123,
        verbose: bool = False,
        **kwargs,
    ):
        super(GSVA, self).__init__(
            outdir=outdir,
            gene_sets=gene_sets,
            module="gsva",
            threads=threads,
            verbose=verbose,
        )
        self.data = data
        self.tau = weight
        self.min_size = min_size
        self.max_size = max_size
        self.seed = seed
        self.mx_diff = mx_diff
        self.abs_rnk = abs_rnk
        self.ranking = None
        self.permutation_num = 0
        self._noplot = True
        if kcdf in ["Gaussian", "gaussian"]:
            self.kernel = True
            self.rnaseq = False
        elif kcdf in ["Poisson", "poisson"]:
            self.kernel = True
            self.rnaseq = True
        else:
            self.kernel = False
            self.rnaseq = False

        # self.figsize = figsize
        # self.format = format
        # self.graph_num = int(graph_num)
        # self.seed = seed
        self.ranking = None
        self.permutation_type = "gene_set"
        assert self.min_size > 0

    def load_data(self) -> pd.DataFrame:
        # load data
        data = self._load_data(self.data)
        data = self._check_data(data)
        return data

    def run(self):
        """run entry"""
        assert self.min_size <= self.max_size
        if self._outdir:
            mkdirs(self.outdir)

        self._logger.info("Parsing data files for GSVA.............................")
        # load data
        df = self.load_data()
        # kernel
        if self.kernel:
            if self.rnaseq:
                self._logger.info(
                    "Estimating ECDFs with Poisson kernels. Clip negative values to 0 !"
                )
                df = df.clip(lower=0)
            else:
                self._logger.info("Estimating ECDFs with Gaussian kernels.")
        else:
            self._logger.info("Estimating ECDFs with directly.")
        # save data
        self.data = df
        # normalized samples, and rank
        # filtering out gene sets and build gene sets dictionary
        self._gene_isupper = self.check_uppercase(gene_list=df.index.values)
        gmt = self.load_gmt(gene_list=df.index.values, gmt=self.gene_sets)
        self.gmt = gmt
        self._logger.info(
            "%04d gene_sets used for further statistical testing....." % len(gmt)
        )
        # start analysis
        self._logger.info("Start to run GSVA...Might take a while................")
        gene_names = df.index.to_list()
        if (not self._gene_isupper) and self._gene_toupper:
            gene_names = [str(x).upper() for x in gene_names]
            self._logger.info("Genes are converted to uppercase.")
        # run
        gsum = gsva_rs(
            gene_names,
            df.values.tolist(),
            gmt,
            self.kernel,
            self.rnaseq,
            self.mx_diff,
            self.abs_rnk,
            self.tau,
            self.min_size,
            self.max_size,
            self._threads,
        )
        self.to_df(gsum.summaries, gmt, df, gsum.indices)
        self.ranking = gsum.rankings
        self._logger.info("Done")



import argparse as ap
import os
import sys

# ------------------------------------
# Main function
# ------------------------------------

# there is a bug in add_argument(required=True), for hacking, don't set metavar='' when required=True,
# or args = argparser.parse_args() will throw bugs!!!


__version__ = "1.1.12"


def main():
    """The Main function/pipeline for GSEApy."""

    # Parse options...
    argparser = prepare_argparser()
    args = argparser.parse_args()
    subcommand = args.subcommand_name

    if subcommand == "replot":
        # reproduce plots using GSEAPY
        from .gsea import Replot

        Replot(
            indir=args.indir,
            outdir=args.outdir,
            weight=args.weight,
            figsize=args.figsize,
            format=args.format,
            verbose=args.verbose,
        ).run()

    elif subcommand == "gsea":
        # compute using GSEAPY
        from .gsea import GSEA

        gs = GSEA(
            args.data,
            args.gmt,
            args.cls,
            args.outdir,
            args.mins,
            args.maxs,
            args.n,
            args.weight,
            args.type,
            args.method,
            args.ascending,
            args.threads,
            args.figsize,
            args.format,
            args.graph,
            args.noplot,
            args.seed,
            args.verbose,
        )
        gs.run()
    elif subcommand == "prerank":
        from .gsea import Prerank

        pre = Prerank(
            args.rnk,
            args.gmt,
            args.outdir,
            args.label[0],
            args.label[1],
            args.mins,
            args.maxs,
            args.n,
            args.weight,
            args.ascending,
            args.threads,
            args.figsize,
            args.format,
            args.graph,
            args.noplot,
            args.seed,
            args.verbose,
        )
        pre.run()

    elif subcommand == "ssgsea":
        from .ssgsea import SingleSampleGSEA

        ss = SingleSampleGSEA(
            data=args.data,
            gene_sets=args.gmt,
            outdir=args.outdir,
            sample_norm_method=args.norm,
            correl_norm_type=args.correl,
            min_size=args.mins,
            max_size=args.maxs,
            permutation_num=args.n,
            weight=args.weight,
            ascending=args.ascending,
            threads=args.threads,
            figsize=args.figsize,
            format=args.format,
            graph_num=args.graph,
            no_plot=args.noplot,
            seed=args.seed,
            verbose=args.verbose,
        )
        ss.run()

    elif subcommand == "gsva":
        from .gsva import GSVA

        gv = GSVA(
            data=args.data,
            gene_sets=args.gmt,
            outdir=args.outdir,
            kcdf=args.kcdf,
            weight=args.weight,
            mx_diff=args.mx_diff,
            abs_rnk=args.abs_rnk,
            min_size=args.mins,
            max_size=args.maxs,
            threads=args.threads,
            seed=args.seed,
            verbose=args.verbose,
        )
        gv.run()
    elif subcommand == "enrichr":
        # calling enrichr API
        from .enrichr_v0 import Enrichr

        enr = Enrichr(
            gene_list=args.gene_list,
            gene_sets=args.library,
            organism=args.organism,
            outdir=args.outdir,
            format=args.format,
            cutoff=args.thresh,
            background=args.bg,
            figsize=args.figsize,
            top_term=args.term,
            no_plot=args.noplot,
            verbose=args.verbose,
        )
        # set organism
        enr.set_organism()
        enr.run()
    elif subcommand == "biomart":
        from .biomart import Biomart

        # read input file or a argument
        name, value = args.filter
        if os.path.isfile(value):
            with open(value, "r") as val:
                lines = val.readlines()
            value = [l.strip() for l in lines]
        # run query
        bm = Biomart(host=args.host, verbose=args.verbose)
        bm.query(
            dataset=args.bg,
            attributes=args.attrs.split(","),
            filters={name: value},
            filename=args.ofile,
        )
    else:
        argparser.print_help()
        sys.exit(0)


def prepare_argparser():
    """Prepare argparser object. New options will be added in this function first."""
    description = "%(prog)s -- Gene Set Enrichment Analysis in Python"
    epilog = "For command line options of each command, type: %(prog)s COMMAND -h"

    # top-level parser
    argparser = ap.ArgumentParser(description=description, epilog=epilog)
    argparser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__
    )
    subparsers = argparser.add_subparsers(
        dest="subcommand_name"
    )  # help="sub-command help")

    # command for 'gsea'
    add_gsea_parser(subparsers)
    # command for 'prerank'
    add_prerank_parser(subparsers)
    # command for 'ssgsea'
    add_singlesample_parser(subparsers)
    # command for 'gsva'
    add_gsva_parser(subparsers)
    # command for 'plot'
    add_plot_parser(subparsers)
    # command for 'enrichr'
    add_enrichr_parser(subparsers)
    # command for 'biomart'
    add_biomart_parser(subparsers)

    return argparser


def add_output_option(parser):
    """output option"""

    parser.add_argument(
        "-o",
        "--outdir",
        dest="outdir",
        type=str,
        default="GSEApy_reports",
        metavar="",
        action="store",
        help="The GSEApy output directory. Default: the current working directory",
    )
    parser.add_argument(
        "-f",
        "--format",
        dest="format",
        type=str,
        metavar="",
        action="store",
        choices=("pdf", "png", "jpeg", "eps", "svg"),
        default="pdf",
        help="File extensions supported by Matplotlib active backend,\
                              choose from {'pdf', 'png', 'jpeg','ps', 'eps','svg'}. Default: 'pdf'.",
    )
    parser.add_argument(
        "--fs",
        "--figsize",
        action="store",
        nargs=2,
        dest="figsize",
        metavar=("width", "height"),
        type=float,
        default=(6.5, 6),
        help="The figsize keyword argument need two parameters to define. Default: (6.5, 6)",
    )
    parser.add_argument(
        "--graph",
        dest="graph",
        action="store",
        type=int,
        default=20,
        metavar="int",
        help="Numbers of top graphs produced. Default: 20",
    )
    parser.add_argument(
        "--no-plot",
        action="store_true",
        dest="noplot",
        default=False,
        help="Speed up computing by suppressing the plot output."
        + "This is useful only if data are interested. Default: False.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        dest="verbose",
        help="Increase output verbosity, print out progress of your job",
    )


def add_output_group(parser, required=True):
    """output group"""

    # output_group = parser.add_mutually_exclusive_group(required=required)
    # output_group.add_argument(
    #     "-o",
    #     "--ofile",
    #     dest="ofile",
    #     type=str,
    #     default="GSEApy_reports",
    #     help="Output file name. Mutually exclusive with --o-prefix.",
    # )
    # output_group.add_argument(
    #     "--o-prefix",
    #     dest="ofile",
    #     type=str,
    #     default="GSEApy_reports",
    #     help="Output file prefix. Mutually exclusive with -o/--ofile.",
    # )
    parser.add_argument(
        "-o",
        "--outdir",
        dest="outdir",
        type=str,
        default="GSEApy_reports",
        metavar="",
        action="store",
        help="The GSEApy output directory. Default: the current working directory",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        dest="verbose",
        help="Increase output verbosity, print out progress of your job",
    )


def add_gsea_parser(subparsers):
    """Add main function 'gsea' argument parsers."""

    argparser_gsea = subparsers.add_parser(
        "gsea", help="Main GSEApy Function: run GSEApy instead of GSEA."
    )

    # group for input files
    group_input = argparser_gsea.add_argument_group("Input files arguments")
    group_input.add_argument(
        "-d",
        "--data",
        dest="data",
        action="store",
        type=str,
        required=True,
        help="Input gene expression dataset file in txt format.Same with GSEA.",
    )
    group_input.add_argument(
        "-c",
        "--cls",
        dest="cls",
        action="store",
        type=str,
        required=True,
        help="Input class vector (phenotype) file in CLS format. Same with GSEA.",
    )
    group_input.add_argument(
        "-g",
        "--gmt",
        dest="gmt",
        action="store",
        type=str,
        required=True,
        help="Gene set database in GMT format. Same with GSEA.",
    )
    group_input.add_argument(
        "-t",
        "--permu-type",
        action="store",
        dest="type",
        type=str,
        metavar="perType",
        choices=("gene_set", "phenotype"),
        default="phenotype",
        help="Type of permutation reshuffling, "
        + "Choose from {'phenotype': 'sample.labels' , 'gene_set' : gene.labels}. "
        + "Default: phenotype",
    )

    # group for output files
    group_output = argparser_gsea.add_argument_group("Output arguments")
    add_output_option(group_output)

    # group for General options.
    group_opt = argparser_gsea.add_argument_group("GSEA advanced arguments")
    group_opt.add_argument(
        "-n",
        "--permu-num",
        dest="n",
        action="store",
        type=int,
        default=1000,
        metavar="nperm",
        help="Number of random permutations. For calculating esnulls. Default: 1000",
    )
    group_opt.add_argument(
        "--min-size",
        dest="mins",
        action="store",
        type=int,
        default=15,
        metavar="int",
        help="Min size of input genes presented in Gene Sets. Default: 15",
    )
    group_opt.add_argument(
        "--max-size",
        dest="maxs",
        action="store",
        type=int,
        default=500,
        metavar="int",
        help="Max size of input genes presented in Gene Sets. Default: 500",
    )
    group_opt.add_argument(
        "-w",
        "--weight",
        action="store",
        dest="weight",
        default=1.0,
        type=float,
        metavar="float",
        help="Weighted_score of rank_metrics. For weighting input genes. Choose from {0, 1, 1.5, 2}. Default: 1",
    )
    group_opt.add_argument(
        "-m",
        "--method",
        action="store",
        dest="method",
        type=str,
        metavar="",
        choices=(
            "signal_to_noise",
            "s2n",
            "abs_signal_to_noise",
            "abs_s2n",
            "t_test",
            "ratio_of_classes",
            "diff_of_classes",
            "log2_ratio_of_classes",
        ),
        default="signal_to_noise",
        help="Methods to calculate correlations of ranking metrics. \
                           Choose from {'signal_to_noise', 'abs_signal_to_noise', 't_test', 'ratio_of_classes', 'diff_of_classes','log2_ratio_of_classes'}.\
                           Default: 'signal_to_noise'",
    )
    group_opt.add_argument(
        "-a",
        "--ascending",
        action="store_true",
        dest="ascending",
        default=False,
        help="Rank metric sorting order. If the -a flag was chosen, then ascending equals to True. Default: False.",
    )
    group_opt.add_argument(
        "-s",
        "--seed",
        dest="seed",
        action="store",
        type=int,
        default=123,
        metavar="",
        help="Number of random seed. Default: 123",
    )
    group_opt.add_argument(
        "-p",
        "--threads",
        dest="threads",
        action="store",
        type=int,
        default=4,
        metavar="procs",
        help="Number of threads you are going to use. Default: 4",
    )

    return


def add_prerank_parser(subparsers):
    """Add function 'prerank' argument parsers."""

    argparser_prerank = subparsers.add_parser(
        "prerank", help="Run GSEApy Prerank tool on preranked gene list."
    )

    # group for input files
    prerank_input = argparser_prerank.add_argument_group("Input files arguments")
    prerank_input.add_argument(
        "-r",
        "--rnk",
        dest="rnk",
        action="store",
        type=str,
        required=True,
        help="Ranking metric file in .rnk format. Same with GSEA.",
    )
    prerank_input.add_argument(
        "-g",
        "--gmt",
        dest="gmt",
        action="store",
        type=str,
        required=True,
        help="Gene set database in GMT format. Same with GSEA.",
    )
    prerank_input.add_argument(
        "-l",
        "--label",
        action="store",
        nargs=2,
        dest="label",
        metavar=("pos", "neg"),
        type=str,
        default=("Pos", "Neg"),
        help="The phenotype label argument need two parameters to define. Default: ('Pos','Neg')",
    )

    # group for output files
    prerank_output = argparser_prerank.add_argument_group("Output arguments")
    add_output_option(prerank_output)

    # group for General options.
    prerank_opt = argparser_prerank.add_argument_group("GSEA advanced arguments")
    prerank_opt.add_argument(
        "-n",
        "--permu-num",
        dest="n",
        action="store",
        type=int,
        default=1000,
        metavar="nperm",
        help="Number of random permutations. For calculating esnulls. Default: 1000",
    )
    prerank_opt.add_argument(
        "--min-size",
        dest="mins",
        action="store",
        type=int,
        default=15,
        metavar="int",
        help="Min size of input genes presented in Gene Sets. Default: 15",
    )
    prerank_opt.add_argument(
        "--max-size",
        dest="maxs",
        action="store",
        type=int,
        default=500,
        metavar="int",
        help="Max size of input genes presented in Gene Sets. Default: 500",
    )
    prerank_opt.add_argument(
        "-w",
        "--weight",
        action="store",
        dest="weight",
        default=1.0,
        type=float,
        metavar="float",
        help="Weighted_score of rank_metrics. For weighting input genes. Choose from {0, 1, 1.5, 2}. Default: 1",
    )
    prerank_opt.add_argument(
        "-a",
        "--ascending",
        action="store_true",
        dest="ascending",
        default=False,
        help="Rank metric sorting order. If the -a flag was chosen, then ascending equals to True. Default: False.",
    )
    prerank_opt.add_argument(
        "-s",
        "--seed",
        dest="seed",
        action="store",
        type=int,
        default=123,
        metavar="",
        help="Number of random seed. Default: 123",
    )
    prerank_opt.add_argument(
        "-p",
        "--threads",
        dest="threads",
        action="store",
        type=int,
        default=4,
        metavar="procs",
        help="Number of threads you are going to use. Default: 4",
    )

    return


def add_singlesample_parser(subparsers):
    """Add function 'singlesample' argument parsers."""

    argparser_gsea = subparsers.add_parser("ssgsea", help="Run Single Sample GSEA.")

    # group for input files
    group_input = argparser_gsea.add_argument_group("Input files arguments")
    group_input.add_argument(
        "-d",
        "--data",
        dest="data",
        action="store",
        type=str,
        required=True,
        help="Input gene expression dataset file in txt format. Same with GSEA.",
    )
    group_input.add_argument(
        "-g",
        "--gmt",
        dest="gmt",
        action="store",
        type=str,
        required=True,
        help="Gene set database in GMT format. Same with GSEA.",
    )
    # group for output files
    group_output = argparser_gsea.add_argument_group("Output arguments")
    add_output_option(group_output)

    # group for General options.
    group_opt = argparser_gsea.add_argument_group(
        "Single Sample GSEA advanced arguments"
    )
    group_opt.add_argument(
        "--sn",
        "--sample-norm",
        dest="norm",
        action="store",
        type=str,
        default="rank",
        metavar="normalize",
        choices=("rank", "log", "log_rank", "custom"),
        help="Sample normalization method. Choose from {'rank', 'log', 'log_rank','custom'}. Default: rank",
    )

    group_opt.add_argument(
        "-c",
        "--correl-type",
        dest="correl",
        action="store",
        type=str,
        default="rank",
        metavar="transform",
        choices=("rank", "symrank", "zscore"),
        help="Input data transformation after sample normalization. Choose from {'rank','symrank', 'zscore'}. Default: rank",
    )
    group_opt.add_argument(
        "--ns",
        "--no-scale",
        action="store_false",
        dest="scale",
        default=True,
        help="If the flag was set, don't normalize the enrichment scores by number of genes.",
    )
    group_opt.add_argument(
        "-n",
        "--permu-num",
        dest="n",
        action="store",
        type=int,
        default=0,
        metavar="nperm",
        help="Number of random permutations. For calculating esnulls. Default: 0",
    )
    group_opt.add_argument(
        "--min-size",
        dest="mins",
        action="store",
        type=int,
        default=15,
        metavar="int",
        help="Min size of input genes presented in Gene Sets. Default: 15",
    )
    group_opt.add_argument(
        "--max-size",
        dest="maxs",
        action="store",
        type=int,
        default=2000,
        metavar="int",
        help="Max size of input genes presented in Gene Sets. Default: 2000",
    )
    group_opt.add_argument(
        "-w",
        "--weight",
        action="store",
        dest="weight",
        default=0.25,
        type=float,
        metavar="float",
        help="Weighted_score of rank_metrics. For weighting input genes. Default: 0.25",
    )
    group_opt.add_argument(
        "-a",
        "--ascending",
        action="store_true",
        dest="ascending",
        default=False,
        help="Rank metric sorting order. If the -a flag was chosen, then ascending equals to True. Default: False.",
    )
    group_opt.add_argument(
        "-s",
        "--seed",
        dest="seed",
        action="store",
        type=int,
        default=123,
        metavar="",
        help="Number of random seed. Default: 123",
    )
    group_opt.add_argument(
        "-p",
        "--threads",
        dest="threads",
        action="store",
        type=int,
        default=4,
        metavar="procs",
        help="Number of Processes you are going to use. Default: 4",
    )

    return


def add_gsva_parser(subparsers):
    """Add function 'GSVA' argument parsers."""

    argparser_gsva = subparsers.add_parser("gsva", help="Run GSVA.")

    # group for input files
    group_input = argparser_gsva.add_argument_group("Input files arguments")
    group_input.add_argument(
        "-d",
        "--data",
        dest="data",
        action="store",
        type=str,
        required=True,
        help="Input gene expression dataset file in txt format. Same with GSEA.",
    )
    group_input.add_argument(
        "-g",
        "--gmt",
        dest="gmt",
        action="store",
        type=str,
        required=True,
        help="Gene set database in GMT format. Same with GSEA.",
    )
    # group for output files
    group_output = argparser_gsva.add_argument_group("Output arguments")
    # add_output_option(group_output)
    add_output_group(group_output)
    # group for General options.
    group_opt = argparser_gsva.add_argument_group("GSVA advanced arguments")
    group_opt.add_argument(
        "-m",
        "--mx-diff",
        dest="mx_diff",
        action="store_false",
        default=True,
        help="When set, ES is calculated as the maximum distance of the random walk from 0. Default: False",
    )

    group_opt.add_argument(
        "-k",
        "--kernel-cdf",
        dest="kcdf",
        action="store",
        type=str,
        default="Gaussian",
        metavar="",
        choices=("Gaussian", "Poisson", "None"),
        help="Gaussian is suitable when input expression values are continuous. "
        + "If input integer counts, then this argument should be set to 'Poisson'",
    )

    group_opt.add_argument(
        "-a",
        "--abs-ranking",
        action="store_true",
        dest="abs_rnk",
        default=False,
        help="Flag used only when --mx-diff is not set. When set, the original Kuiper statistic is used",
    )
    group_opt.add_argument(
        "--min-size",
        dest="mins",
        action="store",
        type=int,
        default=15,
        metavar="int",
        help="Min size of input genes presented in Gene Sets. Default: 15",
    )
    group_opt.add_argument(
        "--max-size",
        dest="maxs",
        action="store",
        type=int,
        default=2000,
        metavar="int",
        help="Max size of input genes presented in Gene Sets. Default: 2000",
    )
    group_opt.add_argument(
        "-w",
        "--weight",
        action="store",
        dest="weight",
        default=1,
        type=float,
        metavar="float",
        help="tau in the random walk performed by the gsva. Default: 1",
    )
    group_opt.add_argument(
        "-s",
        "--seed",
        dest="seed",
        action="store",
        type=int,
        default=123,
        metavar="",
        help="Number of random seed. Default: 123",
    )
    group_opt.add_argument(
        "-p",
        "--threads",
        dest="threads",
        action="store",
        type=int,
        default=4,
        metavar="int",
        help="Number of Processes you are going to use. Default: 4",
    )

    return


def add_plot_parser(subparsers):
    """Add function 'plot' argument parsers."""

    argparser_replot = subparsers.add_parser(
        "replot", help="Reproduce GSEA desktop output figures."
    )

    group_replot = argparser_replot.add_argument_group("Input arguments")

    group_replot.add_argument(
        "-i",
        "--indir",
        action="store",
        dest="indir",
        required=True,
        metavar="GSEA_dir",
        help="The GSEA desktop results directroy that you want to reproduce the figure ",
    )
    add_output_option(group_replot)
    # add_output_group( argparser_plot )
    group_replot.add_argument(
        "-w",
        "--weight",
        action="store",
        dest="weight",
        default=1.0,
        type=float,
        metavar="float",
        help="Weighted_score of rank_metrics. Please Use the same value in GSEA. Choose from (0, 1, 1.5, 2),default: 1",
    )

    return


def add_enrichr_parser(subparsers):
    """Add function 'enrichr' argument parsers."""

    argparser_enrichr = subparsers.add_parser(
        "enrichr", help="Using Enrichr API to perform GO analysis."
    )

    # group for required options.
    enrichr_opt = argparser_enrichr.add_argument_group("Input arguments")
    enrichr_opt.add_argument(
        "-i",
        "--input-list",
        action="store",
        dest="gene_list",
        type=str,
        required=True,
        metavar="IDs",
        help="Enrichr uses a list of gene names as input.",
    )
    enrichr_opt.add_argument(
        "-g",
        "--gene-sets",
        action="store",
        dest="library",
        type=str,
        required=True,
        metavar="GMT",
        help="Enrichr library name(s) required. Separate each name by comma.",
    )
    enrichr_opt.add_argument(
        "--org",
        "--organism",
        action="store",
        dest="organism",
        type=str,
        default="human",
        help="Enrichr supported organism name. Default: human. See here: https://amp.pharm.mssm.edu/modEnrichr.",
    )
    enrichr_opt.add_argument(
        "--ds",
        "--description",
        action="store",
        dest="descrip",
        type=str,
        default="enrichr",
        metavar="STRING",
        help="It is recommended to enter a short description for your list so that multiple lists \
                              can be differentiated from each other if you choose to save or share your list.",
    )
    enrichr_opt.add_argument(
        "--cut",
        "--cut-off",
        action="store",
        dest="thresh",
        metavar="float",
        type=float,
        default=0.05,
        help="Adjust-Pval cutoff, used for generating plots. Default: 0.05.",
    )
    enrichr_opt.add_argument(
        "--bg",
        "--background",
        action="store",
        dest="bg",
        default=None,
        metavar="BG",
        help="Choose background from one of the following. \n"
        + "(1) A BioMart Dataset name, e.g. 'hsapiens_gene_ensembl' . \n"
        + "(2) A total gene number, e.g. 20000. Only works for GMT file input. \n"
        + "(3) A text file contains the background gene list (one gene per row). Gene identifier should be the same to your input (-i). \n"
        + "(4) Default: None. It means all genes in the (-g) input as the background.",
    )
    enrichr_opt.add_argument(
        "-t",
        "--top-term",
        dest="term",
        action="store",
        type=int,
        default=10,
        metavar="int",
        help="Numbers of top terms shown in the plot. Default: 10",
    )
    # enrichr_opt.add_argument("--scale", dest = "scale", action="store", type=float, default=0.5, metavar='float',
    #                          help="scatter dot scale in the dotplot. Default: 0.5")
    # enrichr_opt.add_argument("--no-plot", action='store_true', dest='no_plot', default=False,
    #                           help="Suppress the plot output.This is useful only if data are interested. Default: False.")

    enrichr_output = argparser_enrichr.add_argument_group("Output figure arguments")
    add_output_option(enrichr_output)
    return


def add_biomart_parser(subparsers):
    """Add function 'biomart' argument parsers."""

    argparser_biomart = subparsers.add_parser(
        "biomart", help="Using BioMart API to convert gene ids."
    )

    # group for required options.
    biomart_opt = argparser_biomart.add_argument_group("Input arguments")
    biomart_opt.add_argument(
        "-f",
        "--filter",
        action="store",
        nargs=2,
        dest="filter",
        required=True,
        metavar=("NAME", "VALUE"),
        help="""Which filter to use. Input filter name, and value.
                                     If multi-value required, separate each value by comma.
                                     If value is a txt file, then one ID per row, exclude header.""",
    )
    biomart_opt.add_argument(
        "-a",
        "--attributes",
        action="store",
        dest="attrs",
        type=str,
        required=True,
        metavar="ATTR",
        help="Which attribute(s) to retrieve. Separate each attr by comma.",
    )
    biomart_opt.add_argument(
        "-o", "--ofile", dest="ofile", type=str, required=True, help="Output file name"
    )
    biomart_opt.add_argument(
        "-d",
        "--dataset",
        action="store",
        dest="bg",
        type=str,
        default="hsapiens_gene_ensembl",
        metavar="DATA",
        help="Which dataset to use. Default: hsapiens_gene_ensembl",
    )
    biomart_opt.add_argument(
        "--host",
        action="store",
        dest="host",
        type=str,
        default="www.ensembl.org",
        metavar="HOST",
        help="Which host to use. Select from {'www.ensembl.org', 'asia.ensembl.org', 'useast.ensembl.org'}.",
    )
    biomart_opt.add_argument(
        "-m",
        "--mart",
        action="store",
        dest="mart",
        type=str,
        metavar="MART",
        default="ENSEMBL_MART_ENSEMBL",
        help="Which mart to use. Default: ENSEMBL_MART_ENSEMBL.",
    )
    biomart_opt.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        dest="verbose",
        help="Increase output verbosity, print out progress of your job",
    )


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.stderr.write("User interrupted me! ;-) Bye!\n")
        sys.exit(0)



import logging
import os
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from io import StringIO
from typing import Dict, List, Optional
from xml.etree import cElementTree as ET

import pandas as pd
import requests

from gseapy.utils import log_init, mkdirs, retry


class Biomart:
    """query from BioMart"""

    CHUNK_SIZE = 300  # Default chunk size for batching queries
    MAX_WORKERS: Optional[int] = None  # Default to auto; set an int to override

    def __init__(self, host: str = "www.ensembl.org", verbose: bool = False):
        """simple API to BioMart services.

        How to query validated dataset, attributes, filters.
        Example::
        >>> from gseapy import Biomart
        >>> bm = Biomart(verbose=False, host="ensembl.org")
        >>> ## view validated marts
        >>> marts = bm.get_marts()
        >>> ## view validated dataset
        >>> datasets = bm.get_datasets(mart='ENSEMBL_MART_ENSEMBL')
        >>> ## view validated attributes
        >>> attrs = bm.get_attributes(dataset='hsapiens_gene_ensembl')
        >>> ## view validated filters
        >>> filters = bm.get_filters(dataset='hsapiens_gene_ensembl')
        >>> ## query results
        >>> queries = ['ENSG00000125285','ENSG00000182968'] # a python list
        >>> results = bm.query(dataset='hsapiens_gene_ensembl',
                            attributes=['entrezgene_id', 'go_id'],
                            filters={'ensembl_gene_id': queries}
                            )
        """
        self._id = str(id(self))
        self._logger = log_init(
            name="Biomart" + self._id,
            log_level=logging.INFO if verbose else logging.WARNING,
            filename=None,
        )
        self._set_host(host)

        self.attributes_xml = []
        self.filters_xml = []
        self.dataset_xml = ""

        params = {
            "host": self.host,
            "version": "1.0",
            "virtualSchemaName": "default",
            "formatter": "TSV",
            "header": 0,
            "uniqueRows": 1,
            "configVersion": "0.6",
            "completionStamp": 0,
        }

        self.header = (
            """https://%(host)s/biomart/martservice?query="""
            """<?xml version="%(version)s" encoding="UTF-8"?>"""
            """<!DOCTYPE Query>"""
            """<Query virtualSchemaName="%(virtualSchemaName)s" formatter="%(formatter)s" """
            """header="%(header)s" uniqueRows="%(uniqueRows)s" count="" """
            """datasetConfigVersion="%(configVersion)s" completionStamp="%(completionStamp)s">"""
        )

        self.header = self.header % params
        self.footer = "</Dataset></Query>"
        self.reset()

        # get supported marts
        self._marts = None

    def __del__(self):
        handlers = self._logger.handlers[:]
        for handler in handlers:
            handler.close()  # close file
            self._logger.removeHandler(handler)

    def _set_host(self, host: str):
        """set host"""

        hosts = ["useast.ensembl.org", "asia.ensembl.org"]
        hosts.insert(0, host)
        secure = "s"
        # if host not work, select next
        i = 0
        while i < len(hosts):
            url = "http{}://{}/biomart/martservice?type=registry".format(
                secure, hosts[i]
            )
            request = requests.get(url)
            # '<html>\n\n<head>\n  <title>Service unavailable</title>\n
            # "\n<MartRegistry>\n"
            if request.ok and request.text.startswith("\n<MartRegistry>\n"):
                self.host = hosts[i]
                self._marts = self._get_mart(request.text)
                break
            self._logger.warning(
                "host {} is not reachable, try {} ".format(
                    hosts[i], hosts[(i + 1) % len(hosts)]
                )
            )
            i += 1
        if i == len(hosts):
            self._logger.warning("hosts is not reachable. Please try again later.")

    def add_filter(self, name: str, value: Iterable[str]):
        """
        key: filter names
        value: Iterable[str]
        """
        if isinstance(value, str):
            pass  # keep string as is
        elif isinstance(value, Iterable):
            value = ",".join([str(v) for v in list(value)])
        _filter = ""
        if name.lower().startswith("with"):
            _filter = """<Filter name="%s" excluded="%s"/>""" % (name, value)
        else:
            _filter = """<Filter name="%s" value="%s"/>""" % (name, value)
        self.filters_xml.append(_filter)

    def add_attribute(self, attribute: str):
        _attr = """<Attribute name="%s"/>""" % attribute
        self.attributes_xml.append(_attr)

    def add_dataset(self, dataset: str):
        self.dataset_xml = """<Dataset name="%s" interface="default" >""" % dataset

    def reset(self):
        self.attributes_xml = []
        self.filters_xml = []
        self.dataset_xml = ""

    def get_xml(self):
        xml = self.header
        xml += self.dataset_xml
        for line in self.filters_xml:
            xml += line
        for line in self.attributes_xml:
            xml += line
        xml += self.footer
        return xml

    def get_xml_body(self):
        """Return only the XML body without the URL prefix.

        This is suitable for POST requests where the XML is sent in the body
        as the 'query' form field to the biomart endpoint.
        """
        # self.header starts with 'https://{host}/biomart/martservice?query=' then the XML header
        # Split once on 'query=' to strip the URL and keep the XML portion intact
        try:
            xml_header = self.header.split("query=", 1)[1]
        except Exception:
            # Fallback: if format changes in the future, best effort to keep original behavior
            xml_header = self.header
        xml = xml_header
        xml += self.dataset_xml
        for line in self.filters_xml:
            xml += line
        for line in self.attributes_xml:
            xml += line
        xml += self.footer
        return xml

    def _build_xml_strings(
        self,
        dataset: str,
        attributes: List[str],
        filters: Dict[str, Iterable[str]],
    ) -> Dict[str, str]:
        """Build XML strings without mutating instance state.

        Returns a dict with keys:
          - body: XML suitable for POST body (no URL prefix)
          - url: Full GET URL with query parameter
        """
        # Build dataset and tags locally to keep thread-safety
        dataset_xml = f'<Dataset name="{dataset}" interface="default" >'

        filter_tags: List[str] = []
        for name, value in filters.items():
            if isinstance(value, str):
                v = value
            elif isinstance(value, Iterable):
                v = ",".join([str(vv) for vv in list(value)])
            else:
                v = str(value)
            if name.lower().startswith("with"):
                filter_tags.append(f'<Filter name="{name}" excluded="{v}"/>')
            else:
                filter_tags.append(f'<Filter name="{name}" value="{v}"/>')

        attribute_tags = [f'<Attribute name="{a}"/>' for a in attributes]

        # Header body portion (post-'query=') derived from self.header
        try:
            xml_header = self.header.split("query=", 1)[1]
        except Exception:
            xml_header = self.header
        body = (
            xml_header
            + dataset_xml
            + "".join(filter_tags)
            + "".join(attribute_tags)
            + self.footer
        )
        url = f"https://{self.host}/biomart/martservice?query=" + body
        return {"body": body, "url": url}

    def _get_mart(self, text: str):
        """
        Parse the xml text and return a dataframe of supported marts.

        Parameters
        ----------
        text : str
            a xml text

        Returns
        -------
        marts : pd.DataFrame
            a dataframe of supported marts with columns:
                - Mart: the name of mart
                - Version: the version of mart
        """
        marts = [e.attrib for e in ET.XML(text)]
        marts = pd.DataFrame(marts)
        required_columns = ["database", "displayName", "name"]
        missing = [col for col in required_columns if col not in marts.columns]
        if missing:
            raise ValueError(
                f"BioMart registry XML missing columns: {missing}. Schema may have changed."
            )
        marts = marts.loc[:, required_columns]
        marts.columns = ["Version", "DisplayName", "Mart"]
        # get supported marts
        return marts.loc[:, ["Mart", "Version"]]

    def get_marts(self):
        """Get available marts and their names."""
        url = "https://{host}/biomart/martservice?type=registry&requestid=gseapy{i}".format(
            host=self.host, i=self._id
        )
        if self._marts is not None:
            return self._marts
        resp = requests.get(url)
        if resp.ok and resp.text.startswith("\n<MartRegistry>\n"):
            self._marts = self._get_mart(resp.text)
            return self._marts

        return resp.text

    def get_datasets(self, mart: str = "ENSEMBL_MART_ENSEMBL"):
        """Get available datasets from mart you've selected"""

        marts = self.get_marts()
        if mart not in marts["Mart"].values:
            raise ValueError(
                "Provided mart name (%s) is not valid. see 'names' attribute" % mart
            )

        url = "https://{host}/biomart/martservice?type=datasets&mart={mart}".format(
            host=self.host, mart=mart
        )
        resp = requests.get(url)
        if resp.ok:
            if resp.text.startswith("Problem"):
                return resp.text
            datasets = [
                record.split("\t")
                for record in resp.text.split("\n")
                if len(record) > 1
            ]
            datasets = pd.DataFrame(datasets).iloc[:, 1:3]
            datasets.columns = ["Dataset", "Description"]
            return datasets
        return resp.text

    def get_attributes(self, dataset: str = "hsapiens_gene_ensembl"):
        """Get available attritbutes from dataset you've selected"""
        # assert dataset in

        url = "https://{host}/biomart/martservice?type=attributes&dataset={dataset}".format(
            host=self.host, dataset=dataset
        )
        resp = requests.get(url)
        if resp.ok:
            attributes = [text.split("\t") for text in resp.text.strip().split("\n")]
            attributes = pd.DataFrame(attributes).iloc[:, :3]
            attributes.columns = ["Attribute", "Description", "Additional"]
            return attributes
        return resp.text

    def get_filters(self, dataset: str = "hsapiens_gene_ensembl"):
        """Get available filters from dataset you've selected"""
        # filters = super(Biomart, self).filters(dataset)
        # if dataset not in [x for k in self.valid_attributes.keys() for x in self.valid_attributes[k]]:
        #     raise ValueError("provided dataset (%s) is not found. see valid_attributes" % dataset)
        url = (
            "https://{host}/biomart/martservice?type=filters&dataset={dataset}".format(
                host=self.host, dataset=dataset
            )
        )

        resp = requests.get(url)
        if resp.ok:
            if str(resp.text).startswith("Query ERROR"):
                return resp.text
            filters = [text.split("\t") for text in resp.text.strip().split("\n")]
            df_filters = pd.DataFrame(filters)
            # Check if there are enough columns before selecting
            expected_indices = [0, 1, 3, 5]
            if df_filters.shape[1] >= max(expected_indices) + 1:
                df_filters = df_filters.iloc[:, expected_indices]
                df_filters.columns = [
                    "Filter",
                    "Description",
                    "Additional",
                    "InputType",
                ]
                return df_filters
            else:
                self._logger.warning(
                    f"Filter response has {df_filters.shape[1]} columns, expected at least {max(expected_indices) + 1}. Returning raw DataFrame."
                )
                return df_filters
        return resp.text

    def query(
        self,
        dataset: str = "hsapiens_gene_ensembl",
        attributes: Optional[List[str]] = None,
        filters: Optional[Dict[str, Iterable[str]]] = None,
        filename: Optional[str] = None,
    ):
        """mapping ids using BioMart.

        :param dataset: str, default: 'hsapiens_gene_ensembl'
        :param attributes: str, list, tuple
        :param filters: dict, {'filter name': list(filter value)}
        :param host: www.ensembl.org, asia.ensembl.org, useast.ensembl.org
        :return: a dataframe contains all attributes you selected.

        Example:

            >>> queries = {'ensembl_gene_id': ['ENSG00000125285','ENSG00000182968'] } # need to be a python dict
            >>> results = bm.query(dataset='hsapiens_gene_ensembl',
                                   attributes=['ensembl_gene_id', 'external_gene_name', 'entrezgene_id', 'go_id'],
                                   filters=queries)
        """
        if attributes is None or not attributes:
            attributes = [
                "ensembl_gene_id",
                "external_gene_name",
                "entrezgene_id",
                "go_id",
            ]
        if isinstance(attributes, str):
            attributes = attributes.split(",")

        if filters is None:
            filters = {}
        if not isinstance(filters, dict):
            raise ValueError("filters only accept a dict object")

        df = self.query_simple(
            dataset=dataset, filters=filters, attributes=attributes, filename=None
        )
        if df is None:
            return
        elif isinstance(df, str):
            print(df)
        if "entrezgene_id" in df.columns:
            if not pd.api.types.is_integer_dtype(df["entrezgene_id"]):
                df["entrezgene_id"] = df["entrezgene_id"].astype(pd.Int32Dtype())
        if "entrezgene_id" in df.columns:
            df["entrezgene_id"] = df["entrezgene_id"].astype(pd.Int32Dtype())

        self.results = df
        # save file to cache path.
        if filename is not None:
            mkdirs(os.path.dirname(filename))
            df.to_csv(filename, sep="\t", index=False)

        return df

    def query_simple(
        self,
        dataset: str = "hsapiens_gene_ensembl",
        attributes: Optional[List[str]] = None,
        filters: Optional[Dict[str, Iterable[str]]] = None,
        filename: Optional[str] = None,
    ):
        """
        This function is a simple version of BioMart REST API.
        same parameter to query().

        However, you could get cross page of mapping. such as Mouse 2 human gene names

        **Note**: it will take a couple of minutes to get the results.
        A xml template for querying biomart. (see https://gist.github.com/keithshep/7776579)

        Example::
            >>> from gseapy import Biomart
            >>> bm = Biomart()
            >>> results = bm.query_simple(dataset='mmusculus_gene_ensembl',
                                          attributes=['ensembl_gene_id',
                                                      'external_gene_name',
                                                      'hsapiens_homolog_associated_gene_name'])
        """
        if attributes is None:
            attributes = []
        if filters is None:
            filters = {}

        self.reset()
        self.add_dataset(dataset)
        self._logger.debug("Add attributes")
        for at in attributes:
            self.add_attribute(at)
        self._logger.debug("Add filters")
        for n, v in filters.items():
            self.add_filter(n, v)
        self._logger.debug("Build xml")
        # Build XML for both GET (URL) and POST (body)
        self._xml = self.get_xml()
        xml_body = self.get_xml_body()

        endpoint = f"https://{self.host}/biomart/martservice"
        s = retry(num=5)

        def _parse_response_text(text: str):
            if str(text).startswith("Query ERROR"):
                self._logger.error(text)
                return None
            df_local = pd.read_table(StringIO(text), header=None)
            # Validate column count before assigning names
            if len(df_local.columns) == len(attributes):
                df_local.columns = attributes
            else:
                self._logger.warning(
                    f"Response column count ({len(df_local.columns)}) does not match attributes ({len(attributes)})."
                )
            return df_local

        def _try_post(xml_body_local: str):
            try:
                resp = s.post(endpoint, data={"query": xml_body_local})
            except Exception as e:
                self._logger.warning(f"POST request failed: {e}")
                return None, None
            if resp.ok:
                df_local = _parse_response_text(resp.text)
                if df_local is not None:
                    return df_local, None
                return None, resp.text
            return None, resp.text

        def _try_get(xml_url_local: str):
            try:
                resp = s.get(xml_url_local)
            except Exception as e:
                self._logger.warning(f"GET request failed: {e}")
                return None, None
            if resp.ok:
                df_local = _parse_response_text(resp.text)
                if df_local is not None:
                    return df_local, None
                return None, resp.text
            return None, resp.text

        # Strategy:
        # 1) Prefer POST to avoid 414 (URL too long)
        # 2) Fallback to GET for small queries
        # 3) If still failing (e.g., payload too large), batch the largest list-like filter

        # First attempt: POST once with the whole query
        df = None
        err_text = None

        # Heuristic: if URL is short, it's safe to try GET directly too, but POST is fine for both
        df, err_text = _try_post(xml_body)
        if df is None and (err_text is None or "414" in str(err_text)):
            # Try GET if POST didn't return a valid table (network issue) or explicit 414 indicates GET may also fail
            df, err_text = _try_get(self._xml)

        if df is not None:
            if filename is not None:
                df.to_csv(filename, sep="\t", index=False)
            self.results = df
            return df

        # If we reached here, try batching on the largest list-like filter value
        # Identify list-like filters
        list_like_keys = []
        for k, v in filters.items():
            if isinstance(v, (list, tuple, set, pd.Series)):
                list_like_keys.append((k, list(v)))
        if not list_like_keys:
            # No list-like filters to batch; return the last error text
            return err_text if err_text is not None else "BioMart request failed."

        # Choose the largest filter to batch
        key_to_batch, values_to_batch = max(list_like_keys, key=lambda kv: len(kv[1]))
        if len(values_to_batch) == 0:
            return err_text if err_text is not None else "No values to query."

        # Reasonable chunk size that works reliably with BioMart
        dfs: List[pd.DataFrame] = []

        # Determine chunk size; allow runtime override via instance/class attribute if present
        chunk_size = getattr(self, "CHUNK_SIZE", 300)
        # Build chunk specs first (immutable)
        chunks = []
        for i in range(0, len(values_to_batch), chunk_size):
            sub_values = values_to_batch[i : i + chunk_size]
            chunk_filters = {**filters, key_to_batch: sub_values}
            xmls = self._build_xml_strings(dataset, attributes, chunk_filters)
            chunks.append((i // chunk_size + 1, xmls["body"], xmls["url"]))

        def _fetch_chunk(idx: int, body_xml: str, url_xml: str):
            df_chunk, err_text_chunk = _try_post(body_xml)
            if df_chunk is None:
                df_chunk, err_text_chunk = _try_get(url_xml)
            if df_chunk is None:
                self._logger.warning(f"BioMart chunk {idx} failed: {err_text_chunk}")
            return df_chunk

        # Parallelize chunk fetching with a modest concurrency to improve throughput
        # Keep concurrency bounded to be nice to the BioMart service
        max_workers = getattr(self, "MAX_WORKERS", None)
        if max_workers is None:
            # Enable moderate parallelism for very large inputs; otherwise keep sequential
            max_workers = 1 if len(values_to_batch) < 2000 else 8

        if max_workers > 1 and len(chunks) > 1:
            with ThreadPoolExecutor(max_workers=max_workers) as ex:
                futures = [
                    ex.submit(_fetch_chunk, idx, body, url) for idx, body, url in chunks
                ]
                for fut in as_completed(futures):
                    df_chunk = fut.result()
                    if df_chunk is not None:
                        dfs.append(df_chunk)
        else:
            for idx, body, url in chunks:
                df_chunk = _fetch_chunk(idx, body, url)
                if df_chunk is not None:
                    dfs.append(df_chunk)

        if not dfs:
            return (
                err_text
                if err_text is not None
                else "BioMart request failed in all chunks."
            )

        df_all = pd.concat(dfs, ignore_index=True).drop_duplicates()
        if filename is not None:
            df_all.to_csv(filename, sep="\t", index=False)
        self.results = df_all
        return df_all



# -*- coding: utf-8 -*-
import operator
import sys
import warnings
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import numpy as np
import pandas as pd
import scipy.cluster.hierarchy as sch
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.category import UnitData
from matplotlib.colors import Normalize
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.ticker import MaxNLocator

from gseapy.scipalette import SciPalette


class MidpointNormalize(Normalize):
    def __init__(self, vmin=None, vmax=None, vcenter=None, clip=False):
        self.vcenter = vcenter
        Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        # Note also that we must extrapolate beyond vmin/vmax
        x, y = [self.vmin, self.vcenter, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y, left=-np.inf, right=np.inf))

    def inverse(self, value):
        y, x = [self.vmin, self.vcenter, self.vmax], [0, 0.5, 1]
        return np.interp(value, x, y, left=-np.inf, right=np.inf)


def zscore(data2d: pd.DataFrame, axis: Optional[int] = 0):
    """Standardize the mean and variance of the data axis Parameters.

    :param data2d: DataFrame to normalize.
    :param axis: int, Which axis to normalize across. If 0, normalize across rows,
                  if 1, normalize across columns. If None, don't change data

    :Returns: Normalized DataFrame. Normalized data with a mean of 0 and variance of 1
              across the specified axis.

    """
    if axis is None:
        # normalized to mean and std using entire matrix
        # z_scored = (data2d - data2d.values.mean()) / data2d.values.std(ddof=1)
        return data2d
    assert axis in [0, 1]
    z_scored = data2d.apply(
        lambda x: (x - x.mean()) / x.std(ddof=1), axis=operator.xor(1, axis)
    )
    return z_scored


class Heatmap(object):
    def __init__(
        self,
        df: pd.DataFrame,
        z_score: Optional[int] = None,
        title: Optional[str] = None,
        figsize: Tuple[float, float] = (5, 5),
        cmap: Optional[str] = None,
        xticklabels: bool = True,
        yticklabels: bool = True,
        ofname: Optional[str] = None,
        ax: Optional[plt.Axes] = None,
        **kwargs,
    ):
        self.title = "" if title is None else title
        self.figsize = figsize
        self.xticklabels = xticklabels
        self.yticklabels = yticklabels
        self.ofname = ofname
        self.ax = ax

        # scale dataframe
        df = df.astype(float)
        df = zscore(df, axis=z_score)
        df = df.iloc[::-1]
        self.data = df
        self.cbar_title = "Norm.Exp" if z_score is None else "Z-Score"
        self.cmap = cmap
        if cmap is None:
            self.cmap = SciPalette.create_colormap()  # navyblue2darkred
        self._zscore = z_score

    def _skip_ticks(self, labels, tickevery):
        """Return ticks and labels at evenly spaced intervals."""
        n = len(labels)
        if tickevery == 0:
            ticks, labels = [], []
        elif tickevery == 1:
            ticks, labels = np.arange(n) + 0.5, labels
        else:
            start, end, step = 0, n, tickevery
            ticks = np.arange(start, end, step) + 0.5
            labels = labels[start:end:step]
        return ticks, labels

    def _auto_ticks(self, ax, labels, axis):
        transform = ax.figure.dpi_scale_trans.inverted()
        bbox = ax.get_window_extent().transformed(transform)
        size = [bbox.width, bbox.height][axis]
        axis = [ax.xaxis, ax.yaxis][axis]
        (tick,) = ax.xaxis.set_ticks([0])
        fontsize = tick.label1.get_size()
        max_ticks = int(size // (fontsize / 72))
        if max_ticks < 1:
            tickevery = 1
        else:
            tickevery = len(labels) // max_ticks + 1
        return tickevery

    def get_ax(self):
        """
        Return a matplotlib axes object.

        If an axes is already set, return it.  Otherwise, create a new
        figure and axes instance and set it as the current axes.

        Parameters
        ----------
        None

        Returns
        -------
        ax : matplotlib axes
            The current axes object.
        """
        if (self.ax is not None) and isinstance(self.ax, plt.Axes):
            self.fig = self.ax.figure
            return self.ax
        elif hasattr(sys, "ps1") and (self.ofname is None):
            # Working in an interactive environment, create a figure
            # and show it.
            fig = plt.figure(figsize=self.figsize)
        else:
            # Working non-interactively, create a figure but don't show
            # it.
            fig = Figure(figsize=self.figsize)
            canvas = FigureCanvas(fig)
        ax = fig.add_subplot(111)
        self.fig = fig
        return ax

    def draw(self):
        df = self.data
        ax = self.get_ax()
        vmin = np.percentile(df, 2)
        vmax = np.percentile(df, 98)
        if self._zscore is None:
            norm = Normalize(vmin=vmin, vmax=vmax)
            cbar_locator = MaxNLocator(nbins=5, integer=True)
        else:
            norm = MidpointNormalize(vmin=vmin, vmax=vmax, vcenter=0)
            cbar_locator = MaxNLocator(nbins=3, symmetric=True)  # symmetric=True
        matrix = ax.pcolormesh(
            df.values,
            cmap=self.cmap,
            norm=norm,
            rasterized=True,
        )
        xstep = self._auto_ticks(ax, df.columns.values, 0)
        ystep = self._auto_ticks(ax, df.index.values, 1)
        xticks, xlabels = self._skip_ticks(df.columns.values, tickevery=xstep)
        yticks, ylabels = self._skip_ticks(df.index.values, tickevery=ystep)
        ax.set_ylim([0, len(df)])
        ax.set(xticks=xticks, yticks=yticks)
        ax.set_xticklabels(
            xlabels if self.xticklabels else "", fontsize=14, rotation=90
        )
        ax.set_yticklabels(ylabels if self.yticklabels else "", fontsize=14)
        ax.set_title(self.title, fontsize=20, fontweight="bold")
        ax.tick_params(
            axis="both", which="both", bottom=False, top=False, right=False, left=False
        )
        # cax=fig.add_axes([0.93,0.25,0.05,0.20])
        cbar = self.fig.colorbar(matrix, shrink=0.3, aspect=10)  #  ticks=[-1, 0, 1]
        cbar.ax.yaxis.set_tick_params(
            color="white", direction="in", left=True, right=True
        )
        # Add colorbar, make sure to specify tick locations to match desired ticklabels
        cbar.locator = cbar_locator  # LinearLocator(3)
        cbar.update_ticks()
        cbar.ax.set_title(self.cbar_title, loc="left", fontweight="bold")
        for key, spine in cbar.ax.spines.items():
            spine.set_visible(False)
        # cbar = colorbar(matrix)

        for side in ["top", "right", "left", "bottom"]:
            ax.spines[side].set_visible(False)
            # cbar.ax.spines[side].set_visible(False)
        return ax


def heatmap(
    df: pd.DataFrame,
    z_score: Optional[int] = None,
    title: str = "",
    figsize: Tuple[float, float] = (5, 5),
    cmap: Optional[str] = None,
    xticklabels: bool = True,
    yticklabels: bool = True,
    ofname: Optional[str] = None,
    ax: Optional[plt.Axes] = None,
    **kwargs,
):
    """Visualize the dataframe.

    :param df: DataFrame from expression table.
    :param z_score: 0, 1, or None. z_score axis{0, 1}. If None, not scale.
    :param title: figure title.
    :param figsize: heatmap figsize.
    :param cmap: matplotlib colormap. e.g. "RdBu_r".
    :param xticklabels: bool, whether to show xticklabels.
    :param xticklabels: bool, whether to show xticklabels.
    :param ofname: output file name. If None, don't save figure.
    :param ax: matplotlib axes. Default: None.

    :return: ax if ofname is None.

    """

    ht = Heatmap(
        df=df,
        z_score=z_score,
        title=title,
        figsize=figsize,
        cmap=cmap,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
        ofname=ofname,
        ax=ax,
        **kwargs,
    )
    ax = ht.draw()
    if ofname is None:
        return ax
    # canvas.print_figure(ofname, bbox_inches='tight', dpi=300)
    ht.fig.savefig(ofname, bbox_inches="tight", dpi=300)


class GSEAPlot(object):
    def __init__(
        self,
        term: str,
        tag: Sequence[int],
        runes: Sequence[float],
        nes: float,
        pval: float,
        fdr: float,
        rank_metric: Optional[Sequence[float]] = None,
        pheno_pos: str = "",
        pheno_neg: str = "",
        color: Optional[str] = None,
        figsize: Tuple[float, float] = (6, 5.5),
        cmap: str = "seismic",
        ofname: Optional[str] = None,
        ax: Optional[plt.Axes] = None,
        **kwargs,
    ):
        """
        :param term: gene_set name
        :param tag: hit indices of rank_metric.index presented in gene set S.
        :param runes: running enrichment scores.
        :param nes: Normalized enrichment scores.
        :param pval: nominal p-value.
        :param fdr: false discovery rate.
        :param rank_metric: pd.Series for rankings, rank_metric.values.
        :param pheno_pos: phenotype label, positive correlated.
        :param pheno_neg: phenotype label, negative correlated.
        :param figsize: matplotlib figsize.
        :param ofname: output file name. If None, don't save figure.
        :param ax: matplotlib axes. Default: None.
        """
        # dataFrame of ranked matrix scores
        self.color = "#88C544" if color is None else color
        self._x = np.arange(len(runes))
        self.rankings = None
        self._zero_score_ind = None
        self._z_score_label = None
        if rank_metric is not None:
            self.rankings = np.asarray(rank_metric)
            self._zero_score_ind = np.abs(self.rankings).argmin()
            self._z_score_label = "Zero score at " + str(self._zero_score_ind)

        self.RES = np.asarray(runes)

        self.figsize = figsize
        self.term = term
        self.cmap = cmap
        self.ofname = ofname

        self._pos_label = pheno_pos
        self._neg_label = pheno_neg
        self._hit_indices = tag
        self.module = "tmp" if ofname is None else ofname.split(".")[-2]
        if self.module == "ssgsea":
            self._nes_label = "ES: " + "{:.3f}".format(float(nes))
            self._pval_label = "Pval: invliad for ssgsea"
            self._fdr_label = "FDR: invalid for ssgsea"
        else:
            self._nes_label = "NES: " + "{:.3f}".format(float(nes))
            self._pval_label = "Pval: " + "{:.3e}".format(float(pval))
            self._fdr_label = "FDR: " + "{:.3e}".format(float(fdr))

        # output truetype
        plt.rcParams.update({"pdf.fonttype": 42, "ps.fonttype": 42})
        # in most case, we will have many plots, so do not display plots
        # It's also usefull to run this script on command line.

        # GSEA Plots
        if ax is None:
            if hasattr(sys, "ps1") and (self.ofname is None):
                # working inside python console, show figure
                self.fig = plt.figure(figsize=self.figsize, facecolor="white")
            else:
                # If working on command line, don't show figure
                self.fig = Figure(figsize=self.figsize, facecolor="white")
                self._canvas = FigureCanvas(self.fig)
        elif isinstance(ax, plt.Axes):
            self.fig = ax.figure
        else:
            raise ValueError("ax must be matplotlib axes or None")
        self.fig.suptitle(self.term, fontsize=16, wrap=True, fontweight="bold")

    def axes_rank(self, rect):
        """
        rect : sequence of float
               The dimensions [left, bottom, width, height] of the new axes. All
               quantities are in fractions of figure width and height.
        """
        # Ranked Metric Scores Plot
        ax1 = self.fig.add_axes(rect)
        if self.module == "ssgsea":
            ax1.fill_between(self._x, y1=np.log(self.rankings), y2=0, color="#C9D3DB")
            ax1.set_ylabel("log ranked metric", fontsize=16, fontweight="bold")
        else:
            ax1.fill_between(self._x, y1=self.rankings, y2=0, color="#C9D3DB")
            ax1.set_ylabel("Ranked metric", fontsize=16, fontweight="bold")

        ax1.text(
            0.05,
            0.9,
            self._pos_label,
            color="red",
            horizontalalignment="left",
            verticalalignment="top",
            transform=ax1.transAxes,
        )
        ax1.text(
            0.95,
            0.05,
            self._neg_label,
            color="Blue",
            horizontalalignment="right",
            verticalalignment="bottom",
            transform=ax1.transAxes,
        )
        # the x coords of this transformation are data, and the y coord are axes
        trans1 = transforms.blended_transform_factory(ax1.transData, ax1.transAxes)
        ax1.vlines(
            self._zero_score_ind,
            0,
            1,
            linewidth=0.5,
            transform=trans1,
            linestyles="--",
            color="grey",
        )

        hap = self._zero_score_ind / max(self._x)
        if hap < 0.25:
            ha = "left"
        elif hap > 0.75:
            ha = "right"
        else:
            ha = "center"
        ax1.text(
            hap,
            0.5,
            self._z_score_label,
            horizontalalignment=ha,
            verticalalignment="center",
            transform=ax1.transAxes,
            fontsize=14,
        )
        ax1.set_xlabel("Gene Rank", fontsize=16, fontweight="bold")
        ax1.spines["top"].set_visible(False)
        ax1.tick_params(
            axis="both", which="both", top=False, right=False, left=False, labelsize=14
        )
        ax1.locator_params(axis="y", nbins=5)
        ax1.yaxis.set_major_formatter(
            plt.FuncFormatter(lambda tick_loc, tick_num: "{:.1f}".format(tick_loc))
        )

    def axes_hits(self, rect, bottom: bool = False):
        """
        rect : sequence of float
               The dimensions [left, bottom, width, height] of the new axes. All
               quantities are in fractions of figure width and height.
        """
        # gene hits
        ax2 = self.fig.add_axes(rect)
        # the x coords of this transformation are data, and the y coord are axes
        trans2 = transforms.blended_transform_factory(ax2.transData, ax2.transAxes)
        # to make axes shared with same x cooridincates, make the vlines same ranges to x
        ax2.vlines(
            [self._x[0], self._x[-1]],
            0,
            1,
            linewidth=0.5,
            transform=trans2,
            color="white",
            alpha=0,
        )  # alpha 0 to transparency
        # add hits line
        ax2.vlines(
            self._hit_indices, 0, 1, linewidth=0.5, transform=trans2, color="black"
        )

        ax2.tick_params(
            axis="both",
            which="both",
            bottom=bottom,
            top=False,
            right=False,
            left=False,
            labelbottom=bottom,
            labelleft=False,
        )
        if bottom:
            ax2.set_xlabel("Gene Rank", fontsize=16, fontweight="bold")
            ax2.spines["bottom"].set_visible(True)

    def axes_cmap(self, rect):
        """
        rect : sequence of float
               The dimensions [left, bottom, width, height] of the new axes. All
               quantities are in fractions of figure width and height.
        """
        # center color map at midpoint = 0
        mat = self.rankings
        if self.rankings is None:
            mat = self.RES
        vmin = np.percentile(mat.min(), 2)
        vmax = np.percentile(mat.max(), 98)
        midnorm = MidpointNormalize(vmin=vmin, vcenter=0, vmax=vmax)
        # colormap
        ax3 = self.fig.add_axes(rect)
        ax3.pcolormesh(
            mat[np.newaxis, :],
            rasterized=True,
            norm=midnorm,
            cmap=self.cmap,
        )  # cm.coolwarm
        ax3.spines["bottom"].set_visible(False)
        ax3.tick_params(
            axis="both",
            which="both",
            bottom=False,
            top=False,
            right=False,
            left=False,
            labelbottom=False,
            labelleft=False,
        )

    def axes_stat(self, rect):
        """
        rect : sequence of float
               The dimensions [left, bottom, width, height] of the new axes. All
               quantities are in fractions of figure width and height.
        """
        # Enrichment score plot

        ax4 = self.fig.add_axes(rect)
        ax4.plot(self._x, self.RES, linewidth=4, color=self.color)
        ax4.text(0.1, 0.1, self._fdr_label, transform=ax4.transAxes, fontsize=14)
        ax4.text(0.1, 0.2, self._pval_label, transform=ax4.transAxes, fontsize=14)
        ax4.text(0.1, 0.3, self._nes_label, transform=ax4.transAxes, fontsize=14)

        # the y coords of this transformation are data, and the x coord are axes
        trans4 = transforms.blended_transform_factory(ax4.transAxes, ax4.transData)
        ax4.hlines(0, 0, 1, linewidth=1, transform=trans4, color="grey")
        ax4.set_ylabel("Enrichment Score", fontsize=16, fontweight="bold")
        # ax4.set_xlim(min(self._x), max(self._x))
        ax4.tick_params(
            axis="both",
            which="both",
            bottom=False,
            top=False,
            right=False,
            labelbottom=False,
            labelsize=14,
        )
        ax4.locator_params(axis="y", nbins=5)
        # FuncFormatter need two argument, I don't know why. this lambda function used to format yaxis tick labels.
        ax4.yaxis.set_major_formatter(
            plt.FuncFormatter(lambda tick_loc, tick_num: "{:.1f}".format(tick_loc))
        )

        self.ax = ax4

    def add_axes(self):
        """
        Please check matplotlib docs about how to `add_axes` to figure.

        Here is a more flexible way to create a new gseaplot.
        For example, don't show ranking and merge hits and colormap together
        just used:

            self.axes_stat([0.1,0.2,0.8,0.8]) # axes_stat should be called first
            self.axes_cmap([0.1,0.1,0.8,0.1])
            self.axes_hits([0.1,0.1,0.8,0.1])

        """

        left = 0.1
        width = 0.8
        bottom = 0.1
        height = 0

        stat_height_ratio = 0.4
        hits_height_ratio = 0.05
        cmap_height_ratio = 0.05
        rank_height_ratio = 0.3
        ## make stat /hits height ratio const
        if self.rankings is None:
            rank_height_ratio = 0
            cmap_height_ratio = 0
        base = 0.8 / (
            stat_height_ratio
            + hits_height_ratio
            + cmap_height_ratio
            + rank_height_ratio
        )
        # for i, hit in enumerate(self.hits):
        #     height = hits_height_ratio * base
        if self.rankings is not None:
            height = rank_height_ratio * base
            self.axes_rank([left, bottom, width, height])
            bottom += height
            height = cmap_height_ratio * base
            self.axes_cmap([left, bottom, width, height])
            bottom += height
        height = hits_height_ratio * base
        self.axes_hits(
            [left, bottom, width, height], bottom=False if bottom > 0.1 else True
        )
        bottom += height
        height = stat_height_ratio * base
        self.axes_stat([left, bottom, width, height])
        # self.fig.subplots_adjust(hspace=0)
        # self.fig.tight_layout()

    def savefig(self, bbox_inches="tight", dpi=300):
        # if self.ofname is not None:
        if hasattr(sys, "ps1") and (self.ofname is not None):
            self.fig.savefig(self.ofname, bbox_inches=bbox_inches, dpi=dpi)
        elif self.ofname is None:
            return
        else:
            self._canvas.print_figure(self.ofname, bbox_inches=bbox_inches, dpi=300)
        return


def gseaplot(
    term: str,
    hits: Sequence[int],
    nes: float,
    pval: float,
    fdr: float,
    RES: Sequence[float],
    rank_metric: Optional[Sequence[float]] = None,
    pheno_pos: str = "",
    pheno_neg: str = "",
    color: str = "#88C544",
    figsize: Tuple[float, float] = (6, 5.5),
    cmap: str = "seismic",
    ofname: Optional[str] = None,
    **kwargs,
) -> Optional[List[plt.Axes]]:
    """This is the main function for generating the gsea plot.

    :param term: gene_set name
    :param hits: hits indices of rank_metric.index presented in gene set S.
    :param nes: Normalized enrichment scores.
    :param pval: nominal p-value.
    :param fdr: false discovery rate.
    :param RES: running enrichment scores.
    :param rank_metric: pd.Series for rankings, rank_metric.values.
    :param pheno_pos: phenotype label, positive correlated.
    :param pheno_neg: phenotype label, negative correlated.
    :param color: color for RES and hits.
    :param figsize: matplotlib figsize.
    :param ofname: output file name. If None, don't save figure

    return matplotlib.Figure.
    """
    g = GSEAPlot(
        term,
        hits,
        RES,
        nes,
        pval,
        fdr,
        rank_metric,
        pheno_pos,
        pheno_neg,
        color,
        figsize,
        cmap,
        ofname,
    )
    g.add_axes()
    if ofname is None:
        return g.fig.axes
    g.savefig()


class DotPlot(object):
    def __init__(
        self,
        df: pd.DataFrame,
        x: Optional[str] = None,
        y: str = "Term",
        hue: str = "Adjusted P-value",
        dot_scale: float = 5.0,
        x_order: Optional[List[str]] = None,
        y_order: Optional[List[str]] = None,
        thresh: float = 0.05,
        n_terms: int = 10,
        title: str = "",
        ax: Optional[plt.Axes] = None,
        figsize: Tuple[float, float] = (6, 5.5),
        cmap: str = "viridis_r",
        ofname: Optional[str] = None,
        marker: str = "o",
        **kwargs,
    ):
        """Visualize GSEApy Results with categorical scatterplot
        When multiple datasets exist in the input dataframe, the `x` argument is your friend.

        :param df: GSEApy DataFrame results.
        :param x: Categorical variable in `df` that map the x-axis data. Default: None.
        :param y: Categorical variable in `df` that map the y-axis data. Default: Term.
        :param hue: Grouping variable that will produce points with different colors.
                    Can be either categorical or numeric

        :param x_order: bool, array-like list. Default: False.
                        If True, peformed hierarchical_clustering on X-axis.
                        or input a array-like list of `x` categorical levels.

        :param y_order: bool, array-like list. Default: False.
                        If True, peformed hierarchical_clustering on Y-axis.
                        or input a array-like list of `y` categorical levels.

        :param title: Figure title.
        :param thresh: Terms with `column` value < cut-off are shown. Work only for
                    ("Adjusted P-value", "P-value", "NOM p-val", "FDR q-val")
        :param n_terms: Number of enriched terms to show.
        :param dot_scale: float, scale the dot size to get proper visualization.
        :param ax: Matplotlib axes. Default: None.
        :param figsize: tuple, matplotlib figure size, only used when `ax` is None.
        :param cmap: Matplotlib colormap for mapping the `column` semantic.
        :param ofname: Output file name. If None, don't save figure
        :param marker: The matplotlib.markers. See https://matplotlib.org/stable/api/markers_api.html
        """
        self.marker = marker
        self.y = y
        self.x = x
        self.x_order = x_order
        self.y_order = y_order
        self.hue = str(hue)
        self.colname = str(hue)
        self.figsize = figsize
        self.cmap = cmap
        self.ofname = ofname
        self.ax = ax
        self.scale = dot_scale
        self.title = title
        self.n_terms = n_terms
        self.thresh = thresh
        self.data = self.process(df)
        plt.rcParams.update({"pdf.fonttype": 42, "ps.fonttype": 42})

    def isfloat(self, x):
        try:
            float(x)
        except:
            return False
        else:
            return True

    def process(self, df: pd.DataFrame):
        # check if any values in `df[colname]` can't be coerced to floats
        can_be_coerced = df[self.colname].map(self.isfloat).sum()
        if can_be_coerced < df.shape[0]:
            msg = "some value in %s could not be typecast to `float`" % self.colname
            raise ValueError(msg)
        # subset
        mask = df[self.colname] <= self.thresh
        if self.colname in ["Combined Score", "NES", "ES", "Odds Ratio"]:
            mask.loc[:] = True

        df = df.loc[mask]
        if df.shape[0] < 1:
            msg = "Warning: No enrich terms when cutoff = %s" % self.thresh
            raise ValueError(msg)
        self.cbar_title = self.colname
        # clip GSEA lower bounds
        # if self.colname in ["NOM p-val", "FDR q-val"]:
        #     df[self.colname].clip(1e-5, 1.0, inplace=True)
        # sorting the dataframe for better visualization
        colnd = {
            "Adjusted P-value": "FDR",
            "P-value": "Pval",
            "NOM p-val": "Pval",
            "FDR q-val": "FDR",
        }
        ## impute the 0s in pval, fdr for visualization purpose
        if self.colname in ["Adjusted P-value", "P-value", "NOM p-val", "FDR q-val"]:
            # if all values are zeros, raise error
            if not any(df[self.colname].abs() > 0):
                raise ValueError(
                    f"Can not detetermine colormap. All values in {self.colname} are 0s"
                )
            df = df.sort_values(by=self.colname)
            # add a tiny value to 0s
            p = df[self.colname].astype(float)
            p = np.clip(p, np.finfo(float).tiny, None)  # tiny ≈ 2.22e-308
            df[self.colname] = p
            df = df.assign(p_inv=np.log10(1 / df[self.colname].astype(float)))
            _t = colnd[self.colname]
            self.colname = "p_inv"
            self.cbar_title = r"$\log_{10} \frac{1}{ " + _t + " }$"

        # get top terms; sort ascending
        if (
            (self.x is not None)
            and (self.x in df.columns)
            and (not all(df[self.x].map(self.isfloat)))
        ):
            # if x is numeric column
            # get top term of each group
            df = (
                df.groupby(self.x)
                .apply(lambda _x: _x.sort_values(by=self.colname).tail(self.n_terms))
                .reset_index(drop=True)
            )
        else:
            df = df.sort_values(by=self.colname).tail(self.n_terms)  # acending
        # get scatter area
        if df.columns.isin(["Overlap", "Tag %"]).any():
            ol = df.columns[df.columns.isin(["Overlap", "Tag %"])]
            temp = (
                df[ol].squeeze(axis=1).str.split("/", expand=True).astype(int)
            )  # axis=1, in case you have only 1 row
            df = df.assign(Hits_ratio=temp.iloc[:, 0] / temp.iloc[:, 1])
        else:
            df = df.assign(Hits_ratio=1.0)  # if Overlap column missing
        return df

    def _hierarchical_clustering(self, mat, method, metric) -> List[int]:
        # mat.shape -> [n_sample, m_features]
        Y0 = sch.linkage(mat, method=method, metric=metric)
        Z0 = sch.dendrogram(
            Y0,
            orientation="left",
            # labels=mat.index,
            no_plot=True,
            distance_sort="descending",
        )
        idx = Z0["leaves"][::-1]  # reverse the order to make the view better
        return idx

    def get_x_order(
        self, method: str = "single", metric: str = "euclidean"
    ) -> List[str]:
        """See scipy.cluster.hierarchy.linkage()
        Perform hierarchical/agglomerative clustering.
        Return categorical order.
        """
        if hasattr(self.x_order, "__len__"):
            return self.x_order
        mat = self.data.pivot(
            index=self.y,
            columns=self.x,
            values=self.colname,  # [self.colname, "Hits_ratio"],
        ).fillna(0)
        idx = self._hierarchical_clustering(mat.T, method, metric)
        return list(mat.columns[idx])

    def get_y_order(
        self, method: str = "single", metric: str = "euclidean"
    ) -> List[str]:
        """See scipy.cluster.hierarchy.linkage()
        Perform hierarchical/agglomerative clustering.
        Return categorical order.
        """
        if hasattr(self.y_order, "__len__"):
            return self.y_order
        mat = self.data.pivot(
            index=self.y,
            columns=self.x,
            values=self.colname,  # [self.colname, "Hits_ratio"],
        ).fillna(0)
        idx = self._hierarchical_clustering(mat, method, metric)
        return list(mat.index[idx])

    def get_ax(self):
        """
        Return a matplotlib axes object.

        If an axes is already set, return it.  Otherwise, create a new
        figure and axes instance and set it as the current axes.

        Parameters
        ----------
        None

        Returns
        -------
        ax : matplotlib axes
            The current axes object.
        """
        if (self.ax is not None) and isinstance(self.ax, plt.Axes):
            self.fig = self.ax.figure
            return self.ax
        elif hasattr(sys, "ps1") and (self.ofname is None):
            # Working in an interactive environment, create a figure
            # and show it.
            fig = plt.figure(figsize=self.figsize)
        else:
            # Working non-interactively, create a figure but don't show
            # it.
            fig = Figure(figsize=self.figsize)
            canvas = FigureCanvas(fig)
        ax = fig.add_subplot(111)
        self.fig = fig
        return ax

    def set_x(self):
        """
        set x-axis's value
        """
        x = self.x
        xlabel = ""
        # set xaxis values, so you could get dotplot
        if (x is not None) and (x in self.data.columns):
            xlabel = x
        elif "Combined Score" in self.data.columns:
            xlabel = "Combined Score"
            x = xlabel
        elif "Odds Ratio" in self.data.columns:
            xlabel = "Odds Ratio"
            x = xlabel
        elif "NES" in self.data.columns:
            xlabel = "NES"
            x = xlabel
        else:
            # revert back to p_inv
            x = self.colname
            xlabel = self.cbar_title

        return x, xlabel

    def scatter(
        self,
        outer_ring: bool = False,
    ):
        """
        build scatter
        """
        # scatter colormap range
        # df = df.assign(colmap=self.data[self.colname].round().astype("int"))
        # make area bigger to better visualization
        # area = df["Hits_ratio"] * plt.rcParams["lines.linewidth"] * 100
        df = self.data.assign(
            area=(
                self.data["Hits_ratio"] * self.scale * plt.rcParams["lines.markersize"]
            ).pow(2)
        )
        colmap = df[self.colname].astype(int)
        vmin = np.percentile(colmap.min(), 2)
        vmax = np.percentile(colmap.max(), 98)
        # vmin = np.percentile(df.colmap.min(), 2)
        # vmax = np.percentile(df.colmap.max(), 98)
        ax = self.get_ax()
        # if self.x is None:
        x, xlabel = self.set_x()
        y = self.y
        # if x axis is numberic, prettifiy the plot with the numberic order
        if all(df[x].map(self.isfloat)):
            df = df.sort_values(by=x)
        # set x, y order if set
        xunits = UnitData(self.get_x_order()) if self.x_order else None
        yunits = UnitData(self.get_y_order()) if self.y_order else None

        # outer ring
        if outer_ring:
            smax = df["area"].max()
            # TODO:
            # Matplotlib BUG: when setting edge colors,
            # there's the center of scatter could not aligned.
            # Must set backend to TKcario... to fix it
            # Instead, I just add more dots in the plot to get the ring
            blk_sc = ax.scatter(
                x=x,
                y=y,
                s=smax * 1.6,
                edgecolors="none",
                c="black",
                data=df,
                marker=self.marker,
                xunits=xunits,  # set x categorical order
                yunits=yunits,  # set y categorical order
                zorder=0,
            )
            wht_sc = ax.scatter(
                x=x,
                y=y,
                s=smax * 1.3,
                edgecolors="none",
                c="white",
                data=df,
                marker=self.marker,
                xunits=xunits,  # set x categorical order
                yunits=yunits,  # set y categorical order
                zorder=1,
            )
            # data = np.array(rg.get_offsets()) # get data coordinates
        # inner circle
        sc = ax.scatter(
            x=x,
            y=y,
            data=df,
            s="area",
            edgecolors="none",
            c=self.colname,
            cmap=self.cmap,
            vmin=vmin,
            vmax=vmax,
            marker=self.marker,
            xunits=xunits,  # set x categorical order
            yunits=yunits,  # set y categorical order
            zorder=2,
        )
        ax.set_xlabel(xlabel, fontsize=14, fontweight="bold")
        ax.xaxis.set_tick_params(labelsize=14)
        ax.yaxis.set_tick_params(labelsize=16)
        ax.set_axisbelow(True)  # set grid blew other element
        ax.grid(axis="y", zorder=-1)  # zorder=-1.0
        ax.margins(x=0.25)

        # We change the fontsize of minor ticks label
        # ax.tick_params(axis='y', which='major', labelsize=16)
        # ax.tick_params(axis='both', which='minor', labelsize=14)

        # scatter size legend
        # we use the *func* argument to supply the inverse of the function
        # used to calculate the sizes from above. The *fmt* ensures to string you want
        handles, labels = sc.legend_elements(
            prop="sizes",
            num=3,  #
            # fmt="{x:.2f}",
            color="gray",
            func=lambda s: (
                100 * np.sqrt(s) / plt.rcParams["lines.markersize"] / self.scale
            ),
        )
        ax.legend(
            handles,
            labels,
            title="% Genes\nin set",
            bbox_to_anchor=(1.02, 0.9),
            loc="upper left",
            frameon=False,
            labelspacing=2,
        )
        ax.set_title(self.title, fontsize=20, fontweight="bold")
        self.add_colorbar(sc)

        return ax

    def add_colorbar(self, sc):
        """
        :param sc: matplotlib.Scatter
        """
        # colorbar
        # cax = fig.add_axes([1.0, 0.20, 0.03, 0.22])
        cbar = self.fig.colorbar(
            sc,
            shrink=0.25,
            aspect=10,
            anchor=(0.0, 0.2),  # (0.0, 0.2),
            location="right",
            # cax=cax,
        )
        # cbar.ax.tick_params(direction='in')
        cbar.ax.yaxis.set_tick_params(
            color="white", direction="in", left=True, right=True
        )
        cbar.ax.set_title(self.cbar_title, loc="left", fontweight="bold")
        cbar.ax.title.set_position((0, 1.05))  # 1.05 = 5% above the top
        for key, spine in cbar.ax.spines.items():
            spine.set_visible(False)

    def _parse_colors(self, color=None):
        """
        parse colors for groups
        """
        # map color to group
        if isinstance(color, dict):
            return list(color.values())
        # get default color cycle
        if (not isinstance(color, str)) and hasattr(color, "__len__"):
            _colors = list(color)
        else:
            # get current matplotlib color cycle
            prop_cycle = plt.rcParams["axes.prop_cycle"]
            _colors = prop_cycle.by_key()["color"]
        return _colors

    def barh(self, color=None, group=None, ax=None):
        """
        Barplot
        """
        if ax is None:
            ax = self.get_ax()
        x, xlabel = self.set_x()
        bar = self.data.plot.barh(
            x=self.y, y=self.colname, alpha=0.75, fontsize=16, ax=ax
        )
        if self.hue in ["Adjusted P-value", "P-value", "FDR q-val", "NOM p-val"]:
            xlabel = r"$- \log_{10}$ (%s)" % self.hue
        else:
            xlabel = self.hue
        bar.set_xlabel(xlabel, fontsize=16, fontweight="bold")
        bar.set_ylabel("")
        bar.set_title(self.title, fontsize=24, fontweight="bold")
        bar.xaxis.set_major_locator(MaxNLocator(nbins=5, integer=True))
        #
        _colors = self._parse_colors(color=color)
        colors = _colors
        # remove old legend first
        bar.legend_.remove()
        if (group is not None) and (group in self.data.columns):
            num_grp = self.data[group].value_counts(sort=False)
            # set colors for each bar (groupby hue) using full length
            colors = []
            legend_elements = []
            for i, n in enumerate(num_grp):
                # cycle _colors if num_grp > len(_colors)
                c = _colors[i % len(_colors)]
                # group_label
                label = num_grp.index[i]
                # if input color is a dict with keys in group
                if isinstance(color, dict) and label in color:
                    c = color[label]
                # expand the length to match bars
                colors += [c] * n
                ele = Line2D(
                    xdata=[0],
                    ydata=[0],
                    marker="o",
                    color="w",
                    label=label,
                    markerfacecolor=c,
                    markersize=8,
                )
                legend_elements.append(ele)
            # add custom legend
            ax.legend(
                handles=legend_elements,
                loc="upper left",
                title=group,
                bbox_to_anchor=(1.02, 0.5),
                frameon=False,
            )

        # update color of bars
        for j, b in enumerate(ax.patches):
            c = colors[j % len(colors)]
            b.set_facecolor(c)

        # self.adjust_spines(ax, spines=["left", "bottom"])
        for side in ["right", "top"]:
            ax.spines[side].set_visible(False)
        # set ticks
        ax.tick_params(axis="both", which="both", top=False, right=False)
        return ax

    def to_edgelist(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        return two dataframe of nodes, and edges
        """
        num_nodes = len(self.data)
        # build graph
        # G = nx.Graph()
        group_loc = None
        if self.x is not None:
            group_loc = self.data.columns.get_loc(self.x)
        term_loc = self.data.columns.get_loc(self.y)  # "Terms"
        if "Genes" in self.data.columns:
            gene_loc = self.data.columns.get_loc("Genes")
        elif "Lead_genes":
            gene_loc = self.data.columns.get_loc("Lead_genes")
        else:
            raise KeyError("Sorry, could not locate enriched gene list")
        # build graph
        # nodes = []
        nodes = self.data.reset_index(drop=True)
        nodes.index.name = "node_idx"
        genes = self.data.iloc[:, gene_loc].str.split(";")
        # ns_loc = self.data.columns.get_loc("Hits_ratio")
        edge_list = []
        for i in range(num_nodes):
            # nodes.append([i, self.data.iloc[i, term_loc], self.data.iloc[i, ns_loc]])
            # if group_loc is not None:
            #     nodes[-1].append(self.data.iloc[i, group_loc])
            for j in range(i + 1, num_nodes):
                set_i = set(genes.iloc[i])
                set_j = set(genes.iloc[j])
                ov = set_i.intersection(set_j)
                if len(ov) < 1:
                    continue
                jaccard_coefficient = len(ov) / len(set_i.union(set_j))
                overlap_coefficient = len(ov) / min(len(set_i), len(set_j))
                edge = [
                    i,
                    j,
                    self.data.iloc[i, term_loc],
                    self.data.iloc[j, term_loc],
                    jaccard_coefficient,
                    overlap_coefficient,
                    ",".join(ov),
                ]
                edge_list.append(edge)
                # G.add_edge(src,
                # targ,
                # jaccard= jaccard_coefficient,
                # overlap = overlap_coefficient,
                # genes = list(ov))
        edges = pd.DataFrame(
            edge_list,
            columns=[
                "src_idx",
                "targ_idx",
                "src_name",
                "targ_name",
                "jaccard_coef",
                "overlap_coef",
                "overlap_genes",
            ],
        )
        # node_c = ["node_idx", "node_name", "node_size"]
        # if group_loc is not None:
        #    node_c += ["node_group"]
        # nodes = pd.DataFrame(nodes, columns=node_c)
        return nodes, edges


def dotplot(
    df: pd.DataFrame,
    column: str = "Adjusted P-value",
    x: Optional[str] = None,
    y: str = "Term",
    x_order: Union[List[str], bool] = False,
    y_order: Union[List[str], bool] = False,
    title: str = "",
    cutoff: float = 0.05,
    top_term: int = 10,
    size: float = 5,
    ax: Optional[plt.Axes] = None,
    figsize: Tuple[float, float] = (4, 6),
    cmap: str = "viridis_r",
    ofname: Optional[str] = None,
    xticklabels_rot: Optional[float] = None,
    yticklabels_rot: Optional[float] = None,
    marker: str = "o",
    show_ring: bool = False,
    **kwargs,
):
    """Visualize GSEApy Results with categorical scatterplot
    When multiple datasets exist in the input dataframe, the `x` argument is your friend.

    :param df: GSEApy DataFrame results.
    :param column: column name in `df` that map the dot colors. Default: Adjusted P-value.
    :param x: Categorical variable in `df` that map the x-axis data. Default: None.
    :param y: Categorical variable in `df` that map the y-axis data. Default: Term.

    :param x_order: bool, array-like list. Default: False.
                    If True, peformed hierarchical_clustering on X-axis.
                    or input a array-like list of `x` categorical levels.

    :param x_order: bool, array-like list. Default: False.
                    If True, peformed hierarchical_clustering on Y-axis.
                    or input a array-like list of `y` categorical levels.

    :param title: Figure title.
    :param cutoff: Terms with `column` value < cut-off are shown. Work only for
                   ("Adjusted P-value", "P-value", "NOM p-val", "FDR q-val")
    :param top_term: Number of enriched terms to show (based on values in the `column` (colormap)).
    :param size: float, scale the dot size to get proper visualization.
    :param ax: Matplotlib axes.
    :param figsize: tuple, matplotlib figure size, only used when `ax` is None.
    :param cmap: Matplotlib colormap for mapping the `column` semantic.
    :param ofname: Output file name. If None, don't save figure
    :param marker: The matplotlib.markers. See https://matplotlib.org/stable/api/markers_api.html
    :param show_ring bool: Whether to draw outer ring.

    :return: matplotlib.Axes if ofname is None.
             Only terms with `column` <= `cut-off` are plotted.
    """
    if "group" in kwargs:
        warnings.warn("group is deprecated; use x instead", DeprecationWarning, 2)
        return

    dot = DotPlot(
        df=df,
        x=x,
        y=y,
        x_order=x_order,
        y_order=y_order,
        hue=column,
        title=title,
        thresh=cutoff,
        n_terms=int(top_term),
        dot_scale=size,
        ax=ax,
        figsize=figsize,
        cmap=cmap,
        ofname=ofname,
        marker=marker,
    )
    ax = dot.scatter(outer_ring=show_ring)

    if xticklabels_rot:
        for label in ax.get_xticklabels():
            label.set_ha("right")
            label.set_rotation(xticklabels_rot)

    if yticklabels_rot:
        for label in ax.get_yticklabels():
            label.set_ha("right")
            label.set_rotation(yticklabels_rot)

    if ofname is None:
        return ax
    dot.fig.savefig(ofname, bbox_inches="tight", dpi=300)


def ringplot(
    df: pd.DataFrame,
    column: str = "Adjusted P-value",
    x: Optional[str] = None,
    title: str = "",
    cutoff: float = 0.05,
    top_term: int = 10,
    size: float = 5,
    figsize: Tuple[float, float] = (4, 6),
    cmap: str = "viridis_r",
    ofname: Optional[str] = None,
    xticklabels_rot: Optional[float] = None,
    yticklabels_rot: Optional[float] = None,
    marker="o",
    show_ring: bool = True,
    **kwargs,
):
    """ringplot is deprecated, use dotplot instead

    :param df: GSEApy DataFrame results.
    :param x:  Group by the variable in `df` that will produce categorical scatterplot.
    :param column: column name in `df` to map the dot colors. Default: Adjusted P-value
    :param title: figure title
    :param cutoff: terms with `column` value < cut-off are shown. Work only for
                   ("Adjusted P-value", "P-value", "NOM p-val", "FDR q-val")
    :param top_term: number of enriched terms to show.
    :param size: float, scale the dot size to get proper visualization.
    :param figsize: tuple, matplotlib figure size.
    :param cmap: matplotlib colormap for mapping the `column` semantic.
    :param ofname: output file name. If None, don't save figure
    :param marker: the matplotlib.markers. See https://matplotlib.org/stable/api/markers_api.html
    :param show_ring bool: whether to show outer ring.

    :return: matplotlib.Axes. return None if given ofname.
             Only terms with `column` <= `cut-off` are plotted.
    """
    warnings.warn("ringplot is deprecated; use dotplot instead", DeprecationWarning, 2)
    return


def barplot(
    df: pd.DataFrame,
    column: str = "Adjusted P-value",
    group: Optional[str] = None,
    title: str = "",
    cutoff: float = 0.05,
    top_term: int = 10,
    ax: Optional[plt.Axes] = None,
    figsize: Tuple[float, float] = (4, 6),
    color: Union[str, List[str], Dict[str, str]] = "salmon",
    ofname: Optional[str] = None,
    **kwargs,
):
    """Visualize GSEApy Results.
    When multiple datasets exist in the input dataframe, the `group` argument is your friend.

    :param df: GSEApy DataFrame results.
    :param column: column name in `df` to map the x-axis data. Default: Adjusted P-value
    :param group: group by the variable in `df` that will produce bars with different colors.
    :param title: figure title.
    :param cutoff: terms with `column` value < cut-off are shown. Work only for
                   ("Adjusted P-value", "P-value", "NOM p-val", "FDR q-val")
    :param top_term: number of top enriched terms grouped by `hue` are shown.
    :param ax: Matplotlib axes. If None, create a new figure.
    :param figsize: tuple, matplotlib figsize. only used when ax is None.
    :param color: color or list or dict of matplotlib.colors. Must be reconigzed by matplotlib.
                  if dict input, dict keys must be found in the `group`
    :param ofname: output file name. If None, don't save figure

    :return: matplotlib.Axes. return None if given ofname.
             Only terms with `column` <= `cut-off` are plotted.
    """
    dot = DotPlot(
        df=df,
        x=group if group else None,  # x turns into hue in bar
        y="Term",
        hue=column,  # hue turns into x in bar
        title=title,
        thresh=cutoff,
        n_terms=int(top_term),
        figsize=figsize,
        cmap="viridis",  # placeholder only
        ofname=ofname,
        ax=ax,
    )
    if isinstance(color, str):
        color = [color]
    ax = dot.barh(color=color, group=group)

    if ofname is None:
        return ax
    dot.fig.savefig(ofname, bbox_inches="tight", dpi=300)


class TracePlot(object):
    def __init__(
        self,
        terms: List[str],
        tags: List[Sequence[int]],
        runes: List[Sequence[float]],
        rank_metric: Optional[Sequence[float]] = None,
        figsize: Tuple[float, float] = (6, 5.5),
        colors: Union[str, List[str], None] = None,
        legend_kws: Optional[Dict[str, Any]] = None,
        ofname: Optional[str] = None,
        ax: Optional[plt.Axes] = None,
        **kwargs,
    ):
        """
        terms: list of terms/pathways to show
        tags: list of hit indices for each term
        runes: list runing enrichment score for each term
        hits: list of ranks of the overlap genes in pathway/term.
        rank_metric: ranking metric in descending order.
        colors: list of colors for each term/pathway
        legend_kws: kwargs to pass to ax.legend. e.g. `loc`, `bbox_to_achor`.
        ofname: savefig
        ax: matplotlib's Axes.
        """
        # dataFrame of ranked matrix scores
        self.rankings = rank_metric
        self.terms = terms
        self.runes = runes
        self.hits = tags
        self.figsize = figsize
        if isinstance(colors, str):
            colors = [colors]
        self.colors = colors
        self.ofname = ofname
        self.legend_kws = legend_kws

        # self._pos_label = pheno_pos
        # self._neg_label = pheno_neg
        # output truetype
        plt.rcParams.update({"pdf.fonttype": 42, "ps.fonttype": 42})
        # in most case, we will have many plots, so do not display plots
        # It's also usefull to run this script on command line.
        if ax is None:
            # GSEA Plots
            if hasattr(sys, "ps1") and (self.ofname is None):
                # working inside python console, show figure
                self.fig = plt.figure(figsize=self.figsize, facecolor="white")
            else:
                # If working on command line, don't show figure
                self.fig = Figure(figsize=self.figsize, facecolor="white")
                self._canvas = FigureCanvas(self.fig)
        elif isinstance(ax, plt.Axes):
            self.fig = ax.figure
        else:
            raise ValueError("ax must be matplotlib axes or None")
        # self.fig.suptitle(self.term, fontsize=16, wrap=True, fontweight="bold")

    def axes_hits(
        self, tags: Sequence[int], rect: List[float], color="#0033FF", bottom=False
    ):
        """
        hits: 1d array of this
        rect : sequence of float
               The dimensions [left, bottom, width, height] of the new axes. All
               quantities are in fractions of figure width and height.
        """
        # gene hits
        ax2 = self.fig.add_axes(rect)
        # the x coords of this transformation are data, and the y coord are axes
        trans2 = transforms.blended_transform_factory(ax2.transData, ax2.transAxes)
        # align hits to runes
        ax2.vlines(
            [0, len(self.runes[0]) - 1],
            0,
            1,
            linewidth=0.5,
            transform=trans2,
            color="white",
            alpha=0,
        )  # alpha 0 to transparency
        ax2.vlines(tags, 0, 1, linewidth=0.5, transform=trans2, color=color)
        ax2.spines["bottom"].set_visible(True)
        ax2.tick_params(
            axis="both",
            which="both",
            bottom=bottom,
            top=False,
            right=False,
            left=False,
            labelbottom=bottom,
            labelleft=False,
            labelsize=14,
        )
        if bottom:
            ax2.set_xlabel("Gene Rank", fontsize=16, fontweight="bold")

    def axes_stat(self, rect: List[float]):
        """
        rect : sequence of float
               The dimensions [left, bottom, width, height] of the new axes. All
               quantities are in fractions of figure width and height.
        """
        # Enrichment score plot

        ax4 = self.fig.add_axes(rect)
        if self.rankings is not None:
            ax44 = ax4.twinx()  # instantiate a second axes that shares the same x-axis
            ax44.fill_between(
                range(len(self.rankings)),
                y1=self.rankings,
                y2=0,
                color="#C9D3DB",
                zorder=1,
                alpha=0.5,
            )
            ax44.tick_params(axis="y", labelcolor="#808080")
            ax44.set_ylabel(
                "Ranked metric", fontsize=16, fontweight="bold", color="#808080"
            )
        if self.colors is None:
            cycle = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        else:
            cycle = self.colors
        for i, r, term in zip(range(len(self.terms)), self.runes, self.terms):
            ax4.plot(
                range(len(r)),
                r,
                linewidth=2,
                label=term,
                color=cycle[i % len(cycle)],
                zorder=2 + i,
            )  # color=color)

        # the y coords of this transformation are data, and the x coord are axes
        trans4 = transforms.blended_transform_factory(ax4.transAxes, ax4.transData)
        ax4.hlines(0, 0, 1, linewidth=1, transform=trans4, color="grey")
        ax4.set_ylabel("Enrichment Score", fontsize=16, fontweight="bold")
        # ax4.set_xlim(min(self._x), max(self._x))
        ax4.tick_params(
            axis="both",
            which="both",
            bottom=False,
            top=False,
            right=False,
            labelbottom=False,
            labelsize=14,
        )
        ax4.locator_params(axis="y", nbins=5)
        # FuncFormatter need two argument, I don't know why. this lambda function used to format yaxis tick labels.
        ax4.yaxis.set_major_formatter(
            plt.FuncFormatter(lambda tick_loc, tick_num: "{:.1f}".format(tick_loc))
        )
        if isinstance(self.legend_kws, dict):
            ax4.legend(**self.legend_kws)
        else:
            ax4.legend(loc=(0, 1.02))

        if self.rankings is not None:
            self.align_yaxis(ax4, ax44)

    def add_axes(self):
        """
        Please check matplotlib docs about how to `add_axes` to figure.
        """
        left = 0.1
        width = 0.8
        bottom = 0.1
        # height = 0

        stat_height_ratio = 4
        hits_height_ratio = 0.5
        ## make stat /hits height ratio const
        # 0.8 = base*(4 + len(terms)*0.05)
        base = 0.8 / (stat_height_ratio + len(self.terms) * hits_height_ratio)
        # add each hit track
        if self.colors is None:
            cycle = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        else:
            cycle = self.colors

        for i, hit in enumerate(self.hits):
            height = hits_height_ratio * base
            self.axes_hits(
                tags=hit,
                rect=[left, bottom, width, height],
                color=cycle[i % len(cycle)],
                bottom=False if bottom > 0.1 else True,
            )
            bottom += height
        height = stat_height_ratio * base
        # add mountain curve
        self.axes_stat([left, bottom, width, height])
        # self.fig.subplots_adjust(hspace=0)
        # self.fig.tight_layout()

    def savefig(self, ofname: str, bbox_inches: str = "tight", dpi: float = 300):
        # if self.ofname is not None:
        if hasattr(sys, "ps1") and (ofname is not None):
            self.fig.savefig(self.ofname, bbox_inches=bbox_inches, dpi=dpi)
        elif self.ofname is None:
            return
        else:
            self._canvas.print_figure(ofname, bbox_inches=bbox_inches, dpi=300)
        return

    def align_yaxis(self, ax1, ax2):
        """Align zeros of the two axes, zooming them out by same ratio"""
        axes = np.array([ax1, ax2])
        extrema = np.array([ax.get_ylim() for ax in axes])
        tops = extrema[:, 1] / (extrema[:, 1] - extrema[:, 0])
        # Ensure that plots (intervals) are ordered bottom to top:
        if tops[0] > tops[1]:
            axes, extrema, tops = [a[::-1] for a in (axes, extrema, tops)]

        # How much would the plot overflow if we kept current zoom levels?
        tot_span = tops[1] + 1 - tops[0]

        extrema[0, 1] = extrema[0, 0] + tot_span * (extrema[0, 1] - extrema[0, 0])
        extrema[1, 0] = extrema[1, 1] + tot_span * (extrema[1, 0] - extrema[1, 1])
        [axes[i].set_ylim(*extrema[i]) for i in range(2)]


def gseaplot2(
    terms: List[str],
    hits: List[Sequence[int]],
    RESs: List[Sequence[float]],
    rank_metric: Optional[Sequence[float]] = None,
    colors: Optional[Union[str, List[str]]] = None,
    figsize: Tuple[float, float] = (6, 4),
    legend_kws: Optional[Dict[str, Any]] = None,
    ofname: Optional[str] = None,
    **kwargs,
) -> Optional[List[plt.Axes]]:
    """Trace plot for combining multiple terms/pathways into one plot
    :param terms: list of terms to show in trace plot
    :param hits: list of hits indices correspond to each term.
    :param RESs: list of running enrichment scores correspond to each term.
    :param rank_metric: Optional, rankings.
    :param figsize: matplotlib figsize.
    :legend_kws: Optional, contol the location of lengends
    :param ofname: output file name. If None, don't save figure

    return matplotlib.Figure.
    """
    # in case you just input one pathway
    if isinstance(terms, str):
        terms = [terms]
    # make the inputs are legal
    assert (
        hasattr(terms, "__len__")
        and hasattr(hits, "__len__")
        and hasattr(RESs, "__len__")
    )
    assert len(terms) == len(hits) == len(RESs)

    trace = TracePlot(
        terms=list(terms),
        runes=list(RESs),
        tags=list(hits),
        rank_metric=rank_metric,
        colors=colors,
        figsize=figsize,
        ofname=ofname,
        legend_kws=legend_kws,
        **kwargs,
    )
    trace.add_axes()
    if ofname is None:
        return trace.fig.axes
    trace.savefig(ofname)


def enrichment_map(
    df: pd.DataFrame,
    column: str = "Adjusted P-value",
    cutoff: float = 0.05,
    top_term: int = 10,
    **kwargs,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Visualize GSEApy Results.
    Node size corresponds to the percentage of gene overlap in a certain term of interest.
    Colour of the node corresponds to the significance of the enriched terms.
    Edge size corresponds to the number of genes that overlap between the two connected nodes.
    Gray edges correspond to both nodes when it is the only colour edge.
    When there are two different edge colours, red corresponds to positve nodes and blue corresponds to negative nodes.

    :param df: GSEApy DataFrame results.
    :param column: column name in `df` to map the node colors. Default: Adjusted P-value or FDR q-val.
                   choose from ("Adjusted P-value", "P-value", "FDR q-val", "NOM p-val").

    :param group: group by the variable in `df` that will produce bars with different colors.
    :param title: figure title.
    :param cutoff: nodes with `column` value < cut-off are shown. Work only for
                   ("Adjusted P-value", "P-value", "NOM p-val", "FDR q-val")
    :param top_term: number of top enriched terms are selected as nodes.

    :return: tuple of dataframe (nodes, edges)
    """
    # c = column
    if column not in df.columns:
        for c in ["Adjusted P-value", "P-value", "FDR q-val", "NOM p-val"]:
            if c in df:
                column = c
                break

    dot = DotPlot(
        df=df,
        x=None,  # x turns into hue in colors of nodes
        y="Term",  # node
        hue=column,  # node size
        thresh=cutoff,
        n_terms=int(top_term),
    )
    return dot.to_edgelist()



# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import hypergeom


def calc_pvalues(query, gene_sets, background=20000, **kwargs):
    """calculate pvalues for all categories in the graph

    :param set query: set of identifiers for which the p value is calculated
    :param dict gene_sets: gmt file dict after background was set
    :param set background: total number of genes in your annotated database.
    :returns: pvalues
              x: overlapped gene number
              n: length of gene_set which belongs to each terms
              hits: overlapped gene names.


    For 2*2 contingency table:
    =============================================================================
                         |   in  query  |  not in query |    row total
    =>      in gene_set  |        a     |       b       |       a+b
    =>  not in gene_set  |        c     |       d       |       c+d
           column total                                 | a+b+c+d = anno database
    =============================================================================
    background genes number = a + b + c + d.

    Then, in R
        x=a     the number of white balls drawn without replacement
                from an urn which contains both black and white balls.
        m=a+b   the number of white balls in the urn
        n=c+d   the number of black balls in the urn
        k=a+c   the number of balls drawn from the urn

    In Scipy:
    for args in scipy.hypergeom.sf(k, M, n, N, loc=0):
        M: the total number of objects,
        n: the total number of Type I objects.
        k: the random variate represents the number of Type I objects in N drawn
           without replacement from the total population.

    Therefore, these two functions are the same when using parameters from 2*2 table:
    R:     >   phyper(x-1, m, n, k, lower.tail=FALSE)
    Scipy: >>> hypergeom.sf(x-1, m+n, m, k)

    For Odds ratio in Enrichr (see https://maayanlab.cloud/Enrichr/help#background&q=4)

        oddsRatio = (1.0 * x * d) / Math.max(1.0 * b * c, 1)

    where:

        x are the overlapping genes,
        b (m-x) are the genes in the annotated set - overlapping genes,
        c (k-x) are the genes in the input set - overlapping genes,
        d (bg-m-k+x) are the 20,000 genes (or total genes in the background) - genes in the annotated set - genes in the input set + overlapping genes

    """

    query = set(query)
    vals = []
    # background should be all genes in annotated database
    # such as go, kegg et.al.
    if isinstance(background, set):
        bg = len(background)  # total number in your annotated database
        # filter genes that not found in annotated database
        query = query.intersection(background)
    elif isinstance(background, int):
        bg = background
    else:
        raise ValueError("background should be set or int object")
    # number of genes in your query data
    k = len(query)
    # pval
    subsets = sorted(gene_sets.keys())
    for s in subsets:
        category = set(gene_sets.get(s))
        # the categories should be only exist in custom background too
        if isinstance(background, set):
            category = category.intersection(background)
        hits = query.intersection(category)
        x = len(hits)  # overlap hits
        if x < 1:
            continue
        m = len(category)
        # pVal = hypergeom.sf(hitCount-1,popTotal,bgHits,queryTotal)
        # p(X >= hitCounts)
        pval = hypergeom.sf(x - 1, bg, m, k)
        # oddr, pval2 = odds_ratio_calc(bg, k, m, x)
        # expect_count = k*m/bg
        # oddr= x / expect_count
        # oddr= (x*(bg-m))/(m*(k-x)) # thanks to @sreichl.
        # oddr = ((x + 0.5) * (bg - m + 0.5)) / (
        #     (m + 0.5) * (k - x + 0.5)
        # )  # Haldane-Anscombe correction, issue #132
        bu = 0.5  # base up for Haldane-Anscombe correction. When bu=0, the result is exactly equal to Enrichr.
        oddr = ((x + bu) * (bg - m - k + x + bu)) / (
            (m - x + bu) * (k - x + bu)
        )  # issue #237
        vals.append((s, pval, oddr, x, m, hits))

    return zip(*vals)


# def odds_ratio_calc(bg_n, gene_list_n, gene_set_n, overlap_n):
#     """
#     bg_n = number of background genes
#     gene_list_n = number of genes in the gene list (ie query genes)
#     gene_set_n = number of genes in the (corrected by background) gene set (eg pathways/GO terms)
#     overlap_n = number of genes overlapping with between the (corrected by background) gene set and the gene list
#     """

#     # make contingency table
#     table=np.array([[gene_set_n, bg_n-gene_set_n],[overlap_n, gene_list_n-overlap_n]])
#     # perform Fisher's exact test
#     oddsratio, pvalue = fisher_exact(table)
#     # return (inverse) oddsratio
#     return 1/oddsratio, pvalue


def _ecdf(x):
    nobs = len(x)
    return np.arange(1, nobs + 1) / float(nobs)


def fdrcorrection(pvals, alpha=0.05):
    """benjamini hocheberg fdr correction. inspired by statsmodels"""
    # Implement copy from GOATools.
    pvals = np.asarray(pvals)
    pvals_sortind = np.argsort(pvals)
    pvals_sorted = np.take(pvals, pvals_sortind)

    ecdffactor = _ecdf(pvals_sorted)
    reject = pvals_sorted <= ecdffactor * alpha
    if reject.any():
        rejectmax = max(np.nonzero(reject)[0])
        reject[:rejectmax] = True
    pvals_corrected_raw = pvals_sorted / ecdffactor
    pvals_corrected = np.minimum.accumulate(pvals_corrected_raw[::-1])[::-1]
    del pvals_corrected_raw
    pvals_corrected[pvals_corrected > 1] = 1
    pvals_corrected_ = np.empty_like(pvals_corrected)
    pvals_corrected_[pvals_sortind] = pvals_corrected
    del pvals_corrected
    reject_ = np.empty_like(reject)
    reject_[pvals_sortind] = reject
    return reject_, pvals_corrected_


def multiple_testing_correction(ps, alpha=0.05, method="benjamini-hochberg", **kwargs):
    """correct pvalues for multiple testing and add corrected `q` value

    :param ps: list of pvalues
    :param alpha: significance level default : 0.05
    :param method: multiple testing correction method [bonferroni|benjamini-hochberg]
    :returns (q, rej): two lists of q-values and rejected nodes
    """
    # Implement copy from GOATools.
    _p = np.array(ps)
    q = _p.copy()
    rej = _p.copy()
    mask = ~np.isnan(_p)
    p = _p[mask]
    if method == "bonferroni":
        q[mask] = p * len(p)
        rej[mask] = q[mask] < alpha
    elif method == "benjamini-hochberg":
        _rej, _q = fdrcorrection(p, alpha)
        rej[mask] = _rej
        q[mask] = _q
    else:
        raise ValueError(method)
    return q, rej
